r"""
Witt vector rings

This module provides the class :class:`WittVectorRing` of rings of truncated
Witt vectors.

AUTHORS:

- Jacob Dennerlein (2022-11-28): initial version
- Rubén Muñoz-\-Bertrand (2025-02-13, 2025-12-23): major refactoring and
  new features
"""

# ****************************************************************************
#       Copyright (C) 2025 Rubén Muñoz--Bertrand
#                          <ruben.munoz-bertrand@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from collections.abc import Iterator
from itertools import product

from sage.categories.commutative_additive_groups import CommutativeAdditiveGroups
from sage.categories.commutative_rings import CommutativeRings
from sage.categories.fields import Fields
from sage.categories.homset import Hom
from sage.categories.integral_domains import IntegralDomains
from sage.misc.latex import latex
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.morphism import RingHomomorphism, RingMap
from sage.rings.padics.factory import Zp
from sage.rings.padics.witt_vector import (
    WittVector_finotti,
    WittVector_phantom,
    WittVector_pinvertible,
    WittVector_standard,
)
from sage.rings.polynomial.multi_polynomial import MPolynomial
from sage.rings.polynomial.multi_polynomial_ring_base import MPolynomialRing_base
from sage.rings.polynomial.polynomial_element import Polynomial
from sage.rings.polynomial.polynomial_ring import PolynomialRing_generic
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.sets.primes import Primes
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation


def fast_char_p_power(x, n, p=None):
    r"""
    Return `x^n` assuming that `x` lives in a ring of characteristic `p`.

    If `x` is not an element of a ring of characteristic `p`,
    this throws an error.

    EXAMPLES::

        sage: from sage.rings.padics.witt_vector_ring import fast_char_p_power
        sage: t = GF(1913)(33)
        sage: fast_char_p_power(t, 77)
        1371

    ::

        sage: K.<t> = GF(5^3)
        sage: fast_char_p_power(t, 385)
        4*t^2 + 1
        sage: t^385
        4*t^2 + 1

    ::

        sage: A.<x> = K[]
        sage: fast_char_p_power(x + 1, 10)
        x^10 + 2*x^5 + 1

    ::

        sage: B.<u,v> = K[]
        sage: fast_char_p_power(u + v, 1250)
        u^1250 + 2*u^625*v^625 + v^1250
    """
    x_is_Polynomial = isinstance(x, Polynomial)
    x_is_MPolynomial = isinstance(x, MPolynomial)

    if not (x_is_Polynomial or x_is_MPolynomial):
        return x**n
    if x.is_gen():
        return x**n
    if n < 0:
        x = ~x
        n = -n

    P = x.parent()
    if p is None:
        p = P.characteristic()
    base_p_digits = ZZ(n).digits(base=p)

    xn = 1

    for p_exp, digit in enumerate(base_p_digits):
        if digit == 0:
            continue
        inner_term = x**digit
        term_dict = {}
        for e_int_or_tuple, c in inner_term.dict().items():
            power = p**p_exp
            new_c = fast_char_p_power(c, power)
            new_e_tuple = None
            if x_is_Polynomial:  # Then the dict keys are ints
                new_e_tuple = e_int_or_tuple * power
            elif x_is_MPolynomial:  # Then the dict keys are ETuples
                new_e_tuple = e_int_or_tuple.emul(power)
            term_dict[new_e_tuple] = new_c
        term = P(term_dict)
        xn *= term

    return xn


class WittVectorRing(Parent, UniqueRepresentation):
    r"""
    Return the appropriate `p`-typical truncated Witt vector ring.

    INPUT:

    - ``coefficient_ring`` -- commutative ring of coefficients

    - ``prec`` -- integer (default: `1`), length of the truncated Witt
      vectors in the ring

    - ``p`` -- a prime number (default: ``None``); when it is not set, it
      defaults to the characteristic of ``coefficient_ring`` when it is prime.

    - ``algorithm`` -- the name of the algorithm to use for the ring laws
      (default: ``None``); when it is not set, the most adequate algorithm
      is chosen

    Available algorithms are:

    - ``standard`` -- the schoolbook algorithm;

    - ``finotti`` -- Finotti's algorithm; it can be used when the coefficient
      ring has characteristic `p`;

    - ``phantom`` -- computes the ring laws using the phantom components
      using a lift of ``coefficient_ring``, assuming that it is either
      `\GF{q}` for a power `q` of `p`, or a polynomial ring on that field;

    - ``p_invertible`` -- uses some optimisations when `p` is invertible
      in the coefficient ring.

    EXAMPLES::

        sage: WittVectorRing(QQ, p=5)
        Ring of truncated 5-typical Witt vectors of length 1 over
        Rational Field

    ::

        sage: WittVectorRing(GF(3))
        Ring of truncated 3-typical Witt vectors of length 1 over
        Finite Field of size 3

    ::

        sage: WittVectorRing(GF(3)['t'])
        Ring of truncated 3-typical Witt vectors of length 1 over
        Univariate Polynomial Ring in t over Finite Field of size 3

    ::

        sage: WittVectorRing(Qp(7), prec=30, p=5)
        Ring of truncated 5-typical Witt vectors of length 30 over
        7-adic Field with capped relative precision 20

    ::

        sage: p = 5
        sage: W = WittVectorRing(ZZ, prec=3, p=p)
        sage: W
        Ring of truncated 5-typical Witt vectors of length 3 over Integer Ring
        sage: V = W.verschiebung(extend=True)
        sage: WW = V.codomain()
        sage: WW
        Ring of truncated 5-typical Witt vectors of length 4 over Integer Ring
        sage: F = WW.frobenius_morphism()
        sage: w = W.random_element(x=-100, y=100)
        sage: w  # random
        (-13, 20, -33)
        sage: p*w == F(V(w))
        True

    ::

        sage: p = 7
        sage: W = WittVectorRing(GF(p)['x,y'], prec=5)
        sage: W
        Ring of truncated 7-typical Witt vectors of length 5 over Multivariate
        Polynomial Ring in x, y over Finite Field of size 7
        sage: V = W.verschiebung()
        sage: F = W.frobenius_morphism()
        sage: w = W.random_element()
        sage: w  # random
        (-3*x^2 + 2*y^2 - 3*y + 2, -2*x*y - y^2 - y - 2, -x^2 + y^2,
        x^2 + 2*y^2 + 3*x + y, 2*x^2 - 3*x*y + y^2 + y + 1)
        sage: p*w == F(V(w))
        True
        sage: p*w == V(F(w))
        True

    TESTS::

        sage: A = SymmetricGroup(3).algebra(QQ)
        sage: WittVectorRing(A)
        Traceback (most recent call last):
        ...
        TypeError: Symmetric group algebra of order 3 over Rational Field is not a commutative ring

        sage: WittVectorRing(QQ)
        Traceback (most recent call last):
        ...
        ValueError: Rational Field has non-prime characteristic and no prime was supplied

        sage: WittVectorRing(QQ, p=5, algorithm='moon')
        Traceback (most recent call last):
        ...
        ValueError: algorithm must be one of None, 'standard', 'p_invertible', 'finotti', 'phantom'

        sage: W = WittVectorRing(ZZ, p=53)
        sage: type(W)
        <class 'sage.rings.padics.witt_vector_ring.WittVectorRing_standard_with_category'>

        sage: W = WittVectorRing(PolynomialRing(GF(13), 't'))
        sage: type(W)
        <class 'sage.rings.padics.witt_vector_ring.WittVectorRing_phantom_with_category'>

        sage: W = WittVectorRing(PolynomialRing(GF(5), 't,u'))
        sage: type(W)
        <class 'sage.rings.padics.witt_vector_ring.WittVectorRing_finotti_with_category'>
    """
    def __classcall_private__(cls, coefficient_ring, prec=1, p=None, algorithm=None):
        r"""
        Construct the ring of truncated Witt vectors from the parameters.

        TESTS::

            sage: W = WittVectorRing(QQ, p=5)
            sage: W
            Ring of truncated 5-typical Witt vectors of length 1 over Rational Field
        """
        if coefficient_ring not in CommutativeRings():
            raise TypeError(f"{coefficient_ring} is not a commutative ring")
        elif not isinstance(prec, (int, Integer)):
            raise TypeError(f"{prec} is not an integer")
        elif prec <= 0:
            raise ValueError(f"{prec} must be positive")

        prec = Integer(prec)
        char = coefficient_ring.characteristic()

        if p is None:
            if char not in Primes():
                raise ValueError(f"{coefficient_ring} has non-prime "
                                 "characteristic and no prime was supplied")
            p = char
        elif p not in Primes():
            raise ValueError(f"p must be a prime number, here {p} was given")

        match algorithm:
            case None:
                if p == char:
                    if (coefficient_ring in Fields().Finite()
                        or isinstance(coefficient_ring,
                                      PolynomialRing_generic)
                        and coefficient_ring.base()
                            in Fields().Finite()):
                        child = WittVectorRing_phantom
                    else:
                        child = WittVectorRing_finotti
                elif coefficient_ring(p).is_unit():
                    child = WittVectorRing_pinvertible
                else:
                    child = WittVectorRing_standard
            case 'finotti':
                child = WittVectorRing_finotti
            case 'phantom':
                child = WittVectorRing_phantom
            case 'p_invertible':
                child = WittVectorRing_pinvertible
            case 'standard':
                child = WittVectorRing_standard
            case _:
                raise ValueError("algorithm must be one of None, 'standard', "
                                 "'p_invertible', 'finotti', 'phantom'")

        return child.__classcall__(child, coefficient_ring, prec, p)

    def __init__(self, coefficient_ring, prec, prime, algorithm) -> None:
        r"""
        Initialise ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(PolynomialRing(GF(5), 't'), prec=4); W
            Ring of truncated 5-typical Witt vectors of length 4 over Univariate Polynomial Ring in t over Finite Field of size 5
            sage: type(W)
            <class 'sage.rings.padics.witt_vector_ring.WittVectorRing_phantom_with_category'>

            sage: TestSuite(W).run()
        """
        self._coefficient_ring = coefficient_ring
        self._prec = prec
        self._prime = prime

        if prec.is_one() and coefficient_ring in Fields():
            cat = Fields()
        elif prec.is_one() and coefficient_ring in IntegralDomains():
            cat = IntegralDomains()
        else:
            cat = CommutativeRings()

        if coefficient_ring.base_ring() is coefficient_ring:
            base = self
        else:
            base = WittVectorRing(coefficient_ring.base_ring(), prec=prec,
                                  p=prime, algorithm=algorithm)

        names = tuple('V' + x for x in coefficient_ring.variable_names())

        Parent.__init__(self, base=base, category=cat, names=names)

    def __iter__(self) -> Iterator:
        """
        Iterator for truncated Witt vector rings.

        EXAMPLES::

            sage: W = WittVectorRing(GF(3), p=3, prec=2)
            sage: [w for w in W]
            [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1),
            (2, 2)]
        """
        for t in product(self._coefficient_ring, repeat=self._prec):
            yield self(t)

    def _coerce_map_from_(self, S):
        """"
        Check whether there is a coerce map from ``S``.

        EXAMPLES::

            sage: K = GF(7)
            sage: WK = WittVectorRing(K, prec=3)
            sage: W = WittVectorRing(PolynomialRing(K, 't'), prec=3)
            sage: Wf = WittVectorRing(PolynomialRing(K, 't'), prec=3, algorithm='finotti')
            sage: WW = WittVectorRing(PolynomialRing(K, 't,u'), prec=3)
            sage: Wf.has_coerce_map_from(WK)  # indirect doctest
            True
            sage: Wf.has_coerce_map_from(W)  # indirect doctest
            False
            sage: WW.has_coerce_map_from(W)  # indirect doctest
            True
            sage: WW.has_coerce_map_from(Wf)  # indirect doctest
            True
            sage: WW.has_coerce_map_from(WK)  # indirect doctest
            True

            sage: W = WittVectorRing(GF(25), prec=2)
            sage: W.has_coerce_map_from(WittVectorRing(GF(5), prec=3))  # indirect doctest
            True
            sage: W.has_coerce_map_from(WittVectorRing(ZZ, p=5, prec=3))  # indirect doctest
            True
            sage: W.has_coerce_map_from(WittVectorRing(PolynomialRing(GF(5), 't,u'), prec=3))  # indirect doctest
            False
            sage: WW = WittVectorRing(PolynomialRing(GF(5), 't'), prec=3)
            sage: WW.has_coerce_map_from(W)  # indirect doctest
            False
            sage: W.has_coerce_map_from(WW)  # indirect doctest
            False

            sage: W = WittVectorRing(QQ, p=3, prec=3)
            sage: W.has_coerce_map_from(WittVectorRing(ZZ, p=3, prec=3))  # indirect doctest
            True
            sage: W.has_coerce_map_from(WittVectorRing(QQ, p=3, prec=2))  # indirect doctest
            False

            sage: W = WittVectorRing(PolynomialRing(ZZ, 'x'), p=5, prec=2)
            sage: W.has_coerce_map_from(WittVectorRing(ZZ, p=5, prec=3))  # indirect doctest
            True
            sage: W.has_coerce_map_from(WittVectorRing(ZZ, p=3, prec=3))  # indirect doctest
            False
        """
        if (isinstance(S, WittVectorRing)
            and S.precision() >= self._prec and S.prime() == self._prime
            and self._coefficient_ring.has_coerce_map_from(
                S.coefficient_ring())):
            return (any(isinstance(S, rng) for rng in self._always_coerce)
                    or (S.precision() != self._prec
                        or S.coefficient_ring() is not self._coefficient_ring)
                    and any(isinstance(S, rng)
                            for rng in self._coerce_when_different))
        if S is ZZ:
            return True

    def _generate_witt_polynomials(self, coefficient_ring, prec, p):
        """
        Generate the sum and product polynomials defining the ring laws of
        truncated Witt vectors for the ``standard`` algorithm.

        EXAMPLES::

            sage: P.<X1,X2,Y1,Y2> = PolynomialRing(GF(3),'X1,X2,Y1,Y2')
            sage: W = WittVectorRing(P, p=3, prec=2, algorithm='standard')
            sage: W([X1,X2]) + W([Y1,Y2])  # indirect doctest
            (X1 + Y1, -X1^2*Y1 - X1*Y1^2 + X2 + Y2)
            sage: W([X1,X2]) * W([Y1,Y2])  # indirect doctest
            (X1*Y1, X2*Y1^3 + X1^3*Y2)
            sage: W.frobenius_morphism()(W([X1,X2]))  # indirect doctest
            (X1^3, X2^3)
        """
        x_var_names = [f'X{i}' for i in range(prec)]
        y_var_names = [f'Y{i}' for i in range(prec)]
        var_names = x_var_names + y_var_names

        # Okay, what's going on here? Sage, by default, relies on
        # Singular for Multivariate Polynomial Rings, but Singular uses
        # only SIXTEEN bits (unsigned) to store its exponents. So if we
        # want exponents larger than 2^16 - 1, we have to use the
        # generic implementation. However, after some experimentation,
        # it seems like the generic implementation is faster?
        #
        # After trying to compute S_4 for p=5, it looks like generic is
        # faster for  very small polys, and MUCH slower for large polys.
        # So we'll default to singular unless we can't use it.
        #
        # Remark: Since when is SIXTEEN bits sufficient for anyone???
        #
        if p**(prec - 1) >= 2**16:
            implementation = 'generic'
        else:
            implementation = 'singular'

        # We first generate the "universal" polynomials and then project
        # to the coefficient ring.
        R = PolynomialRing(ZZ, var_names, implementation=implementation)
        x_y_vars = R.gens()
        x_vars = x_y_vars[:prec]
        y_vars = x_y_vars[prec:]

        self._sum_polynomials = [0]*(prec)
        for n in range(prec):
            s_n = x_vars[n] + y_vars[n]
            for i in range(n):
                s_n += ((x_vars[i]**(p**(n-i)) + y_vars[i]**(p**(n-i))
                        - self._sum_polynomials[i]**(p**(n-i))) / p**(n-i))
            self._sum_polynomials[n] = R(s_n)

        self._prod_polynomials = [x_vars[0] * y_vars[0]] + [0]*(prec-1)
        for n in range(1, prec):
            x_poly = sum([p**i * x_vars[i]**(p**(n-i)) for i in range(n+1)])
            y_poly = sum([p**i * y_vars[i]**(p**(n-i)) for i in range(n+1)])
            p_poly = sum([p**i * self._prod_polynomials[i]**(p**(n-i))
                         for i in range(n)])
            p_n = (x_poly*y_poly - p_poly) // p**n
            self._prod_polynomials[n] = p_n

        R = PolynomialRing(ZZ, x_var_names, implementation=implementation)
        x_vars = R.gens()

        self._frob_polynomials = []
        if not prec.is_one():
            self._frob_polynomials = [x_vars[0]**p + p*x_vars[1]]
            for n in range(2, prec):
                x_poly = sum([p**i * x_vars[i]**(p**(n-i)) for i in range(n+1)])
                p_poly = sum([p**i * self._frob_polynomials[i]**(p**(n-1-i))
                              for i in range(n-1)])
                self._frob_polynomials.append((x_poly - p_poly) / p**(n-1))

        S = PolynomialRing(coefficient_ring, x_y_vars)
        for n in range(prec):
            self._sum_polynomials[n] = S(self._sum_polynomials[n])
            self._prod_polynomials[n] = S(self._prod_polynomials[n])
        S = PolynomialRing(coefficient_ring, x_vars)
        for n in range(prec-1):
            self._frob_polynomials[n] = S(self._frob_polynomials[n])

    def _latex_(self) -> str:
        r"""
        Return a `\LaTeX` representation of ``self``.

        .. WARNING::

            This representation follows the standard representation in the
            literature which does not mention `p`.

        EXAMPLES::

            sage: W = WittVectorRing(PolynomialRing(GF(3),'t'))
            sage: latex(W)
            W_{1}\left(\Bold{F}_{3}[t]\right)
        """
        return "W_{%s}\\left(%s\\right)" % (latex(self._prec),
                                            latex(self._coefficient_ring))

    def _repr_(self) -> str:
        """
        Return a string representation of the ring.

        EXAMPLES::

            sage: WittVectorRing(QQ, p=2, prec=5)
            Ring of truncated 2-typical Witt vectors of length 5 over Rational Field
        """
        return f"Ring of truncated {self._prime}-typical Witt vectors of "\
               f"length {self._prec} over {self._coefficient_ring}"

    def cardinality(self):
        """
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: WittVectorRing(GF(17), prec=2).cardinality()
            289
            sage: WittVectorRing(QQ, p=2).cardinality()
            +Infinity
        """
        return self._coefficient_ring.cardinality()**(self._prec)

    def characteristic(self):
        """
        Return the characteristic of ``self``.

        EXAMPLES::

            sage: WittVectorRing(GF(25), p=5, prec=3).characteristic()
            125
            sage: WittVectorRing(ZZ, p=2, prec=4).characteristic()
            0
            sage: WittVectorRing(Integers(18), p=3, prec=3).characteristic()
            162
        """
        p = self._prime
        if self._coefficient_ring(p).is_unit():
            # If p is invertible, W_n(R) is isomorphic to R^n.
            return self._coefficient_ring.characteristic()

        # This is Jacob Dennerlein's Corollary 3.3. in "Computational
        # Aspects of Mixed Characteristic Witt Vectors" (preprint)
        return p**(self._prec-1) * self._coefficient_ring.characteristic()

    def coefficient_ring(self):
        """
        Return the coefficient ring of ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(Zp(5), p=5)
            sage: W.coefficient_ring()
            5-adic Ring with capped relative precision 20
        """
        return self._coefficient_ring

    def frobenius_morphism(self, truncate=False):
        """
        Return the Frobenius homomorphism of this ring of Witt vectors.

        INPUT:

        - ``truncate`` -- boolean (default: ``False``); when the coefficient
          ring has positive characteristic `p`, chose whether to lower the
          precision of the codomain by `1`. This is always the case when the
          characteristic of the coefficient ring is not `p`.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ)
            sage: W = WittVectorRing(R, p=5, prec=2)
            sage: w = W([x, y])
            sage: frob = W.frobenius_morphism()
            sage: frob(w)
            (x^5 + 5*y)

        TESTS::

            sage: TestSuite(frob).run(skip="_test_category")
        """
        return WittVectorFrobeniusMorphism(self, truncate=truncate)

    def frobenius_polynomials(self, variables=None):
        """
        Return the Witt Frobenius polynomials.

        INPUT:

        - ``variables`` -- names of the indeterminates (default: ``None``),
          given as a string, or as a list of strings, whose length must be the
          precision of the ring. When nothing is given, variables indexed by
          `X` are used.

        EXAMPLES::

            sage: W = WittVectorRing(GF(5), prec=4)
            sage: W.frobenius_polynomials()
            [X0^5, X1^5, X2^5]

            sage: W = WittVectorRing(ZZ, p=2, prec=3)
            sage: W.frobenius_polynomials('T0, T1, T2')
            [T0^2 + 2*T1, -2*T0^2*T1 - T1^2 + 2*T2]

            sage: W = WittVectorRing(ZZ, p=3, prec=3)
            sage: W.frobenius_polynomials('T0, T1, T2')
            [T0^3 + 3*T1, -3*T0^6*T1 - 9*T0^3*T1^2 - 8*T1^3 + 3*T2]
        """
        if not hasattr(self, '_frob_polynomials'):
            self._generate_witt_polynomials(self._coefficient_ring, self._prec,
                                            self._prime)
        if variables is None:
            return self._frob_polynomials.copy()
        R = PolynomialRing(self._coefficient_ring, variables)
        return [R(self._frob_polynomials[i]) for i in range(self._prec-1)]

    def gen(self, n=0):
        """
        Return the ``n``-th generator of ``self``.

        .. WARNING::

            This is not the ``n``-th generator of ``self`` as a
            ``self.base_ring()``-algebra.

        .. SEEALSO::

            :meth:`gens`

        EXAMPLES::

            sage: R.<x, y> = ZZ[]
            sage: W = WittVectorRing(R, p=3, prec=3)
            sage: W.gen()
            (x, 0, 0)
            sage: W.gen(1)
            (y, 0, 0)
            sage: W.gen(n=2)
            Traceback (most recent call last):
            ...
            ValueError: Generator not defined.
        """
        return self.teichmuller_lift(self._coefficient_ring.gen(n))

    def gens(self) -> tuple:
        r"""
        Return a tuple of generators of ``self``.

        .. WARNING::

            These are not the generators of ``self`` as a
            ``self.base_ring()``-algebra. Rather, this set generates a
            sub-``self.base_ring()``-algebra `A` of ``self`` such that every
            element of ``self`` can be written as `\sum_{i=0}^{prec}V^i(a_i)`
            where all `a_i\in A`.

        EXAMPLES::

            sage: W = WittVectorRing(GF(29), prec=3)
            sage: W.gens()
            ((1, 0, 0),)

            sage: F.<a> = GF(27)
            sage: W = WittVectorRing(F, prec=3)
            sage: W.gens()
            ((a, 0, 0),)

            sage: R.<x> = ZZ[]
            sage: W = WittVectorRing(R, p=7, prec=2)
            sage: W.gens()
            ((x, 0),)
        """
        return tuple(map(self.teichmuller_lift, self._coefficient_ring.gens()))

    def is_exact(self) -> bool:
        """
        Return whether the elements of ``self`` are represented exactly.

        EXAMPLES::

            sage: WittVectorRing(ZZ, p=3, prec=3).is_exact()
            True
            sage: WittVectorRing(Qp(7), p=7, prec=2).is_exact()
            False
        """
        return self._coefficient_ring.is_exact()

    def is_field(self, proof=True):
        """
        Return ``True`` if this ring is a field.

        INPUT:

        - ``proof`` -- boolean (default: ``True``); if set to ``True``, the
          returned value is correct but the method might throw an exception.
          Otherwise, if it is set to ``False``, the method returns ``True`` if
          it can establish that ``self`` is a field, and ``False`` otherwise.

        EXAMPLES::

            sage: WittVectorRing(QQ, p=2).is_field()
            True
            sage: WittVectorRing(QQ, p=2, prec=2).is_field()
            False
            sage: WittVectorRing(GF(9, 'a')).is_field()
            True
            sage: WittVectorRing(GF(9, 'a'), prec=3).is_field()
            False
            sage: WittVectorRing(ZZ, p=5).is_field()
            False
            sage: L.<z> = LazyLaurentSeriesRing(ZZ)
            sage: W = WittVectorRing(L, p=7, algorithm="standard")
            sage: W.is_field(proof=True)
            Traceback (most recent call last):
            ...
            NotImplementedError: unable to determine whether or not Lazy
            Laurent Series Ring in z over Integer Ring is a field.
            sage: W.is_field(proof=False)
            False
        """
        return self._prec.is_one() and self._coefficient_ring.is_field(proof=proof)

    def is_finite(self) -> bool:
        """
        Return whether ``self`` is a finite ring.

        EXAMPLES::

            sage: WittVectorRing(GF(23)).is_finite()
            True
            sage: WittVectorRing(ZZ, p=2).is_finite()
            False
        """
        return self._coefficient_ring.is_finite()

    def is_integral_domain(self, proof=True):
        """
        Return ``True`` if this ring is an integral domain.

        INPUT:

        - ``proof`` -- boolean (default: ``True``); if set to ``True``, the
          returned value is correct but the method might throw an exception.
          Otherwise, if it is set to ``False``, the method returns ``True`` if
          it can establish that ``self`` is an integral domain, and ``False``
          otherwise.

        EXAMPLES::

            sage: WittVectorRing(ZZ, p=2).is_integral_domain()
            True
            sage: WittVectorRing(ZZ, p=2, prec=2).is_integral_domain()
            False
            sage: WittVectorRing(GF(9, 'a')).is_integral_domain()
            True
            sage: WittVectorRing(GF(9, 'a'), prec=3).is_integral_domain()
            False
            sage: R.<x> = ZZ[]
            sage: S.<y> = R.quo(x^2 + 7*x + 8)
            sage: T.<t> = S[]
            sage: Q.<z> = T.quo(t^2)
            sage: W = WittVectorRing(Q, p=7, algorithm="standard")
            sage: W.is_integral_domain(proof=True)
            Traceback (most recent call last):
            ...
            NotImplementedError: cannot rewrite Univariate Quotient
            Polynomial Ring in y over Integer Ring with modulus
            x^2 + 7*x + 8 as an isomorphic ring
            sage: W.is_integral_domain(proof=False)
            False
        """
        return (self._prec.is_one() and
                self._coefficient_ring.is_integral_domain(proof=proof))

    def is_integrally_closed(self):
        """
        Return ``True`` if this ring is an integrally closed domain.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^2 + 189*x + 394)
            sage: R = K.order(2*a)
            sage: WittVectorRing(R, p=79).is_integrally_closed()
            False
            sage: WittVectorRing(ZZ, p=101).is_integrally_closed()
            True
            sage: WittVectorRing(ZZ, p=101, prec=2).is_integrally_closed()
            False
            sage: S.<t> = ZZ[]
            sage: WittVectorRing(S, p=101).is_integrally_closed()
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        return (self._prec.is_one() and
                self._coefficient_ring.is_integrally_closed())

    def is_prime_field(self):
        r"""
        Return ``True`` if ``self`` is isomorphic to one of the prime fields
        `\QQ` or `\GF{p}`.

        EXAMPLES::

            sage: WittVectorRing(QQ, p=2).is_prime_field()
            True
            sage: WittVectorRing(QQ, p=2, prec=2).is_prime_field()
            False
            sage: WittVectorRing(GF(9, 'a')).is_prime_field()
            False
            sage: WittVectorRing(GF(7)).is_prime_field()
            True
            sage: WittVectorRing(GF(7), prec=3).is_prime_field()
            False
        """
        return self._prec.is_one() and self._coefficient_ring.is_prime_field()

    def ngens(self) -> int:
        """
        Return the number of generators of ``self``.

        .. WARNING::

            This is not the number of generators of ``self`` as a
            ``self.base_ring()``-algebra.

        .. SEEALSO::

            :meth:`gens`

        EXAMPLES::

            sage: W = WittVectorRing(GF(2), p=3, prec=3)
            sage: W.ngens()
            1
            sage: W = WittVectorRing(GF(2)['t'], p=3, prec=3)
            sage: W.ngens()
            1
            sage: W = WittVectorRing(GF(3), p=3, prec=3)
            sage: W.ngens()
            1
            sage: W = WittVectorRing(GF(3)['t'], p=3, prec=3)
            sage: W.ngens()
            1
            sage: W = WittVectorRing(GF(9,'a'), p=3, prec=3)
            sage: W.ngens()
            1
            sage: W = WittVectorRing(GF(9,'a')['t'], p=3, prec=3)
            sage: W.ngens()
            1
        """
        return len(self.gens())

    def precision(self):
        """
        Return the length of the truncated Witt vectors in ``self``.

        EXAMPLES::

            sage: WittVectorRing(GF(9), p=3, prec=3).precision()
            3
        """
        return self._prec

    def prime(self):
        """
        Return the prime from which the truncated Witt vector ring has been
        constructed.

        EXAMPLES::

            sage: W = WittVectorRing(GF(81), prec=3)
            sage: W.prime()
            3

            sage: W = WittVectorRing(ZZ, p=7, prec=2)
            sage: W.prime()
            7
        """
        return self._prime

    def prod_polynomials(self, variables=None):
        """
        Return the Witt product polynomials.

        INPUT:

        - ``variables`` -- names of the indeterminates (default: ``None``),
          given as a string, or as a list of strings, whose length must be the
          double of the precision of the ring. When nothing is given,
          variables indexed by `X` and `Y` are used.

        EXAMPLES::

            sage: W = WittVectorRing(GF(5), prec=3)
            sage: W.prod_polynomials()
            [X0*Y0,
            X1*Y0^5 + X0^5*Y1,
            -X0^5*X1^4*Y0^20*Y1 - 2*X0^10*X1^3*Y0^15*Y1^2 - 2*X0^15*X1^2*Y0^10*Y1^3 - X0^20*X1*Y0^5*Y1^4 + X2*Y0^25 + X0^25*Y2 + X1^5*Y1^5]

            sage: W = WittVectorRing(ZZ, p=2, prec=2)
            sage: W.prod_polynomials('T0, T1, U0, U1')
            [T0*U0, T1*U0^2 + T0^2*U1 + 2*T1*U1]
        """
        if not hasattr(self, '_prod_polynomials'):
            self._generate_witt_polynomials(self._coefficient_ring, self._prec,
                                            self._prime)
        if variables is None:
            return self._prod_polynomials.copy()
        R = PolynomialRing(self._coefficient_ring, variables)
        return [R(self._prod_polynomials[i]) for i in range(self._prec)]

    def random_element(self, *args, **kwds):
        """
        Return a random truncated Witt vector.

        Extra arguments are passed to
        the random generator of the coefficient ring.

        EXAMPLES::

            sage: WittVectorRing(GF(27,'t'), prec=2).random_element()  # random
            (z3, 2*z3^2 + 1)

            sage: W = WittVectorRing(PolynomialRing(ZZ,'x'), p=3, prec=3)
            sage: W.random_element(5)  # random
            (x^5 - 2*x^4 - 4*x^3 - 2*x^2 + 1, -x^5 + 2*x^4 - x - 1,
            -x^5 + 7*x^4 + 3*x^3 - 24*x^2 - 1)
        """
        return self(tuple(self._coefficient_ring.random_element(*args, **kwds)
                          for _ in range(self._prec)))

    def sum_polynomials(self, variables=None):
        """
        Return the Witt sum polynomials.

        INPUT:

        - ``variables`` -- names of the indeterminates (default: ``None``),
          given as a string, or as a list of strings, whose length must be the
          double of the precision of the ring. When nothing is given,
          variables indexed by `X` and `Y` are used.

        EXAMPLES::

            sage: W = WittVectorRing(GF(5), prec=2)
            sage: W.sum_polynomials(['T0', 'T1', 'U0', 'U1'])
            [T0 + U0, -T0^4*U0 - 2*T0^3*U0^2 - 2*T0^2*U0^3 - T0*U0^4 + T1 + U1]

            sage: W = WittVectorRing(ZZ, p=2, prec=3)
            sage: W.sum_polynomials()
            [X0 + Y0,
            -X0*Y0 + X1 + Y1,
            -X0^3*Y0 - 2*X0^2*Y0^2 - X0*Y0^3 + X0*X1*Y0 + X0*Y0*Y1 - X1*Y1 + X2 + Y2]
        """
        if not hasattr(self, '_sum_polynomials'):
            self._generate_witt_polynomials(self._coefficient_ring, self._prec,
                                            self._prime)
        if variables is None:
            return self._sum_polynomials.copy()
        R = PolynomialRing(self._coefficient_ring, variables)
        return [R(self._sum_polynomials[i]) for i in range(self._prec)]

    def teichmuller_lift(self, x):
        """
        Return the Teichmüller lift of ``x`` in ``self``.

        This lift is sometimes known as the multiplicative lift of ``x``.

        EXAMPLES::

            sage: WittVectorRing(GF(125,'t'), prec=2).teichmuller_lift(3)
            (3, 0)
        """
        if x not in self._coefficient_ring:
            raise TypeError(f"{x} not in {self._coefficient_ring}")
        return self((x,) + tuple(0 for _ in range(self._prec-1)))

    def verschiebung(self, extend=False):
        """
        Return the Verschiebung map of this ring of Witt vectors.

        INPUT:

        - ``extend`` -- boolean (default: ``False``); whether the codomain of
          the map has precision `1` more than ``self``.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(QQ)
            sage: W = WittVectorRing(R, p=5, prec=2)
            sage: w = W([x, y])
            sage: V = W.verschiebung()
            sage: V(w)
            (0, x)

        TESTS::

            sage: TestSuite(V).run(skip="_test_category")
        """
        return WittVectorVerschiebung(self, extend=extend)


class WittVectorRing_finotti(WittVectorRing):
    """
    Child class for truncated Witt vectors using Finotti's algorithm.

    .. WARNING::

        This class should never be called directly, use ``WittVectorRing``
        instead.

    EXAMPLES::

        sage: W = WittVectorRing(GF(49), prec=3, algorithm='finotti')
        sage: W
        Ring of truncated 7-typical Witt vectors of length 3 over Finite Field in z2 of size 7^2

        sage: W = WittVectorRing(ZZ, p=11, prec=3, algorithm='finotti')
        Traceback (most recent call last):
        ...
        ValueError: the 'finotti' algorithm only works for coefficients rings of characteristic p
    """
    Element = WittVector_finotti

    def __init__(self, coefficient_ring, prec, prime) -> None:
        r"""
        Initialise ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(PowerSeriesRing(GF(3), 't'), p=3, prec=3)
            sage: W
            Ring of truncated 3-typical Witt vectors of length 3 over Power Series Ring in t over Finite Field of size 3
            sage: type(W)
            <class 'sage.rings.padics.witt_vector_ring.WittVectorRing_finotti_with_category'>

            sage: TestSuite(W).run()
        """
        if coefficient_ring.characteristic() != prime:
            raise ValueError("the 'finotti' algorithm only works for "
                             "coefficients rings of characteristic p")

        if isinstance(coefficient_ring, MPolynomialRing_base):
            self._always_coerce = [WittVectorRing_finotti,
                                   WittVectorRing_phantom,
                                   WittVectorRing_standard]
            self._coerce_when_different = []
        else:
            self._always_coerce = [WittVectorRing_finotti,
                                   WittVectorRing_standard]
            self._coerce_when_different = [WittVectorRing_phantom]

        import numpy as np
        R = Zp(prime, prec=prec+1, type='fixed-mod')
        v_p = ZZ.valuation(prime)
        table = [[0]]
        for k in range(1, prec+1):
            pk = prime**k
            row = np.empty(pk, dtype=int)
            row[0] = 0
            prev_bin = 1
            for i in range(1, pk // 2 + 1):
                val = v_p(i)
                # Instead of calling binomial each time, we compute the
                # coefficients recursively. This is MUCH faster.
                next_bin = prev_bin * (pk - (i-1)) // i
                prev_bin = next_bin
                series = R(-next_bin // prime**(k-val))
                for _ in range(val):
                    temp = series % prime
                    series = (series - R.teichmuller(temp)) // prime
                row[i] = ZZ(series % prime)
                row[pk - i] = row[i]  # binomial coefficients are symmetric
            table.append(row)
        self._binomial_table = table

        super().__init__(coefficient_ring, prec, prime, "finotti")

    def _eta_bar(self, vec, eta_index):
        r"""
        Generate the `\eta_i` for ``finotti``'s algorithm.

        EXAMPLES::

            sage: R.<x,y,z,t> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=2, algorithm='finotti')
            sage: (W([x,y]) + W([z,t]))[1]  # indirect doctest
            -x^4*z - 2*x^3*z^2 - 2*x^2*z^3 - x*z^4 + y + t
        """
        vec = tuple(x for x in vec if x != 0)  # strip zeroes

        # special cases
        if len(vec) <= 1:
            return 0
        if eta_index == 0:
            return sum(vec)

        # renaming to match notation in Finotti's "Computations with Witt
        # vectors and the Greenberg transform", doi:10.1142/S1793042114500377
        k = eta_index
        p = self._prime
        # if vec = (x,y), we know what to do: Theorem 8.6
        if len(vec) == 2:
            # Here we have to check if we've pre-computed already
            x, y = vec
            scriptN = [[None] for _ in range(k+1)]  # each list starts with
            # None, so that indexing matches paper

            # calculate first N_t scriptN's
            for t in range(1, k+1):
                for i in range(1, p**t):
                    scriptN[t].append(self._binomial_table[t][i]
                                      * fast_char_p_power(x, i)
                                      * fast_char_p_power(y, p**t - i))
            indexN = [p**i - 1 for i in range(k+1)]
            for t in range(2, k+1):
                for i in range(1, t):
                    # append scriptN_{t, N_t+l}
                    next_scriptN = self._eta_bar(
                        scriptN[t-i][1:indexN[t-i]+t-i], i
                    )
                    scriptN[t].append(next_scriptN)
            return sum(scriptN[k][1:])

        # if vec is longer, we split and recurse: Proposition 5.4
        # This is where we need to using multiprocessing.
        m = len(vec) // 2
        v_1 = vec[:m]
        v_2 = vec[m:]
        s_1 = sum(v_1)
        s_2 = sum(v_2)
        scriptM = [[] for _ in range(k+1)]
        for t in range(1, k+1):
            scriptM[t].append(self._eta_bar(v_1, t))
            scriptM[t].append(self._eta_bar(v_2, t))
            scriptM[t].append(self._eta_bar((s_1, s_2), t))
        for t in range(2, k+1):
            for s in range(1, t):
                result = self._eta_bar(scriptM[t-s], s)
                scriptM[t].append(result)
        return sum(scriptM[k])


class WittVectorRing_phantom(WittVectorRing):
    """
    Child class for truncated Witt vectors using the ``phantom`` algorithm.

    .. WARNING::

        This class should never be called directly, use ``WittVectorRing``
        instead.

    EXAMPLES::

        sage: W = WittVectorRing(GF(19), prec=20)
        sage: W
        Ring of truncated 19-typical Witt vectors of length 20 over Finite Field of size 19

        sage: W = WittVectorRing(QQ, p=23, prec=3, algorithm='phantom')
        Traceback (most recent call last):
        ...
        ValueError: the 'phantom' algorithm only works when the coefficient ring is a finite field of char. p, or a polynomial ring on that field
    """
    Element = WittVector_phantom

    def __init__(self, coefficient_ring, prec, prime) -> None:
        r"""
        Initialise ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(GF(23,'t'), p=23, prec=2)
            sage: W
            Ring of truncated 23-typical Witt vectors of length 2 over Finite Field of size 23
            sage: type(W)
            <class 'sage.rings.padics.witt_vector_ring.WittVectorRing_phantom_with_category'>

            sage: TestSuite(W).run()
        """
        msg = "the 'phantom' algorithm only works when the coefficient ring is"\
            " a finite field of char. p, or a polynomial ring on that field"

        if coefficient_ring.characteristic() != prime:
            raise ValueError(msg)

        if not (coefficient_ring in Fields().Finite() or
                (isinstance(coefficient_ring, (PolynomialRing_generic,
                                               MPolynomialRing_base)) and
                 coefficient_ring.base() in Fields().Finite())):
            raise ValueError(msg)

        if (coefficient_ring in Fields().Finite()
            or isinstance(coefficient_ring,
                          PolynomialRing_generic)):
            self._always_coerce = [WittVectorRing_finotti,
                                   WittVectorRing_phantom,
                                   WittVectorRing_standard]
            self._coerce_when_different = []
        else:
            self._always_coerce = [WittVectorRing_phantom,
                                   WittVectorRing_standard]
            self._coerce_when_different = [WittVectorRing_finotti]

        super().__init__(coefficient_ring, prec, prime, "phantom")


class WittVectorRing_pinvertible(WittVectorRing):
    """
    Child class for truncated Witt vectors using the ``p_invertible`` algorithm.

    .. WARNING::

        This class should never be called directly, use ``WittVectorRing``
        instead.

    EXAMPLES::

        sage: W = WittVectorRing(QQ, p=31, prec=20)
        sage: W
        Ring of truncated 31-typical Witt vectors of length 20 over Rational Field

        sage: W = WittVectorRing(GF(3), prec=3, algorithm='p_invertible')
        Traceback (most recent call last):
        ...
        ValueError: the 'p_invertible' algorithm only works when p is a unit in the ring of coefficients
    """
    Element = WittVector_pinvertible

    def __init__(self, coefficient_ring, prec, prime) -> None:
        r"""
        Initialise ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(QQ, p=11, prec=3)
            sage: W
            Ring of truncated 11-typical Witt vectors of length 3 over Rational Field
            sage: type(W)
            <class 'sage.rings.padics.witt_vector_ring.WittVectorRing_pinvertible_with_category'>

            sage: TestSuite(W).run()
        """
        if not coefficient_ring(prime).is_unit():
            raise ValueError("the 'p_invertible' algorithm only works when p "
                             "is a unit in the ring of coefficients")

        self._always_coerce = [WittVectorRing_pinvertible,
                               WittVectorRing_standard]
        self._coerce_when_different = []

        super().__init__(coefficient_ring, prec, prime, "p_invertible")


class WittVectorRing_standard(WittVectorRing):
    """
    Child class for truncated Witt vectors using the ``standard`` algorithm.

    .. WARNING::

        This class should never be called directly, use ``WittVectorRing``
        instead.

    EXAMPLES::

        sage: W = WittVectorRing(GF(3), prec=3, algorithm='standard')
        sage: W
        Ring of truncated 3-typical Witt vectors of length 3 over Finite Field of size 3
    """
    Element = WittVector_standard

    def __init__(self, coefficient_ring, prec, prime) -> None:
        r"""
        Initialise ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(ZZ, p=5, prec=2)
            sage: W
            Ring of truncated 5-typical Witt vectors of length 2 over Integer Ring
            sage: type(W)
            <class 'sage.rings.padics.witt_vector_ring.WittVectorRing_standard_with_category'>

            sage: TestSuite(W).run()
        """
        self._always_coerce = []
        self._coerce_when_different = [WittVectorRing]

        self._generate_witt_polynomials(coefficient_ring, prec, prime)

        super().__init__(coefficient_ring, prec, prime, "standard")


class WittVectorFrobeniusMorphism(RingHomomorphism):
    """
    A class implementing Frobenius homomorphisms on Witt vector rings.

    .. WARNING::

        This class should not be called directly, use
        :meth:`WittVectorRing.frobenius_morphism` instead.

    EXAMPLES::

        sage: R.<x,y,z> = PolynomialRing(ZZ)
        sage: W = WittVectorRing(R, p=3, prec=3)
        sage: w = W([x, y, z])
        sage: frob = W.frobenius_morphism()
        sage: frob(w)
        (x^3 + 3*y, -3*x^6*y - 9*x^3*y^2 - 8*y^3 + 3*z)

        sage: W = WittVectorRing(R, p=3, prec=1)
        sage: W.frobenius_morphism()
        Traceback (most recent call last):
        ...
        ValueError: the ring of Witt vectors must have precision at least 2
        when truncate=True

    TESTS::

        sage: TestSuite(frob).run(skip="_test_category")
    """
    def __init__(self, domain, truncate=False):
        """
        INPUT:

        - ``domain`` -- the ring of Witt vectors whose Frobenius endomorphism
          we are computing.

        - ``truncate`` -- boolean (default: ``False``); when the coefficient
          ring of the domain has positive characteristic `p`, chose whether to
          lower the precision of the codomain by `1`. This is always the case
          when the characteristic of the coefficient ring is not `p`.

        EXAMPLES::

            sage: R.<x,y,z,t> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=4)
            sage: w = W([x, y, z, t])
            sage: frob = W.frobenius_morphism()
            sage: frob(w)
            (x^5, y^5, z^5, t^5)
            sage: frob = W.frobenius_morphism(truncate=True)
            sage: frob(w)
            (x^5, y^5, z^5)

        TESTS::

            sage: R.<x,y> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=2)
            sage: frob = W.frobenius_morphism()
            sage: TestSuite(frob).run(skip="_test_category")
            sage: frob = W.frobenius_morphism(truncate=True)
            sage: TestSuite(frob).run(skip="_test_category")
        """

        coeff_ring = domain.coefficient_ring()

        if not truncate:
            if coeff_ring.characteristic() != domain.prime():
                truncate = True

            else:
                codomain = domain
                self._call_ = self._call_char_p

        if truncate:
            prec = domain.precision()

            if prec.is_one():
                raise ValueError("the ring of Witt vectors must have "
                                 "precision at least 2 when truncate=True")

            if isinstance(domain, WittVectorRing_finotti):
                algorithm = "finotti"
                self._call_ = self._call_char_p_truncate
            elif isinstance(domain, WittVectorRing_phantom):
                algorithm = "phantom"
                self._call_ = self._call_phantom_truncate
            elif isinstance(domain, WittVectorRing_pinvertible):
                algorithm = "p_invertible"
                self._call_ = self._call_standard
            elif isinstance(domain, WittVectorRing_standard):
                algorithm = "standard"
                self._call_ = self._call_standard

            codomain = WittVectorRing(coeff_ring, prec=prec-1,
                                      p=domain.prime(), algorithm=algorithm)

        super().__init__(Hom(domain, codomain))

    def _call_char_p(self, x):
        """
        Evaluate ``self`` at ``x`` when the coefficient ring has characteristic `p`
        for the non truncated Frobenius.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=3, algorithm="phantom")
            sage: w = W([x, x + y, x*y + y])
            sage: frob = W.frobenius_morphism()
            sage: frob(w)  # indirect doctest
            (x^5, x^5 + y^5, x^5*y^5 + y^5)

        TESTS::

            sage: TestSuite(frob).run(skip="_test_category")
        """
        W = self.domain()
        F = W.coefficient_ring().frobenius_endomorphism()
        return W([F(w) for w in x])

    def _call_char_p_truncate(self, x):
        """
        Evaluate ``self`` at ``x`` when the coefficient ring has charaterisic `p`
        for the truncated Frobenius.

        EXAMPLES::

            sage: W = WittVectorRing(GF(7), prec=4, algorithm="finotti")
            sage: w = W([1, 2, 3, 4])
            sage: frob = W.frobenius_morphism(truncate=True)
            sage: frob(w)  # indirect doctest
            (1, 2, 3)

        TESTS::

            sage: TestSuite(frob).run(skip="_test_category")
        """
        Wcod = self.codomain()
        F = Wcod.coefficient_ring().frobenius_endomorphism()
        return Wcod([F(w) for w in x])

    def _call_phantom_truncate(self, x):
        """
        Evaluate ``self`` at ``x`` when the algorithm is ``phantom`` and the
        Frobenius is truncated.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=3, algorithm="phantom")
            sage: w = W([x, x + y, x*y + y])
            sage: frob = W.frobenius_morphism(truncate=True)
            sage: fw = frob(w)
            sage: fw.phantom(lift=True)  # indirect doctest
            (x^5 + 5*x + 5*y, x^25 + 5*x^5 + 5^2*x^4*y + 2*5^2*x^3*y^2 +
            2*5^2*x^2*y^3 + 5^2*x*y^4 + 5*y^5 + 5^2*x*y + 5^2*y)

        TESTS::

            sage: TestSuite(frob).run(skip="_test_category")
        """
        phantom = x.phantom(lift=True)
        Wcod = self.codomain()
        return Wcod(phantom=phantom[1:])

    def _call_standard(self, x):
        """
        Evaluate ``self`` at ``x`` using the general algorithm.

        EXAMPLES::

            sage: W = WittVectorRing(ZZ, p=11, prec=2)
            sage: w = W([1, 100])
            sage: frob = W.frobenius_morphism()
            sage: frob(w)  # indirect doctest
            (1101)

        TESTS::

            sage: TestSuite(frob).run(skip="_test_category")
        """
        W = self.domain()
        Wcod = self.codomain()
        return Wcod([W.frobenius_polynomials()[i](x.coordinates())
                     for i in range(Wcod.precision())])

    def _latex_(self) -> str:
        r"""
        Return a `\LaTeX` representation of ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(PolynomialRing(GF(3),'t'))
            sage: latex(W.frobenius_morphism())
            F
        """
        return "F"

    def _repr_(self) -> str:
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(GF(11), prec=5)
            sage: W.frobenius_morphism()
            Frobenius endomorphism on the Ring of truncated 11-typical Witt
            vectors of length 5 over Finite Field of size 11
            sage: W.frobenius_morphism(truncate=True)
            Frobenius homomorphism from the Ring of truncated 11-typical Witt
            vectors of length 5 over Finite Field of size 11 to the Ring of
            truncated 11-typical Witt vectors of length 4 over Finite Field of
            size 11
        """
        if self.domain() is self.codomain():
            return f"Frobenius endomorphism on the {self.domain()}"
        return f"Frobenius homomorphism from the {self.domain()} to the "\
               f"{self.codomain()}"


class WittVectorVerschiebung(RingMap):
    """
    A class implementing Veschiebung maps on Witt vector rings.

    .. WARNING::

        This class should not be called directly, use
        :meth:`WittVectorRing.verschiebung` instead.

    EXAMPLES::

        sage: R.<x,y,z> = PolynomialRing(ZZ)
        sage: W = WittVectorRing(R, p=3, prec=3)
        sage: w = W([x, y, z])
        sage: V = W.verschiebung()
        sage: V(w)
        (0, x, y)

    TESTS::

        sage: TestSuite(V).run(skip="_test_category")
    """
    def __init__(self, domain, extend=False):
        """
        INPUT:

        - ``domain`` -- the ring of Witt vectors whose Verschiebung map we are
          computing.

        - ``extend`` -- boolean (default: ``False``); whether the codomain of
          the Verschiebung map is a ring of Witt vectors whose lengths are `1`
          more than the ones in the domain.

        EXAMPLES::

            sage: R.<x,y,z,t> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=4)
            sage: w = W([x, y, z, t])
            sage: V = W.verschiebung()
            sage: V(w)
            (0, x, y, z)
            sage: V = W.verschiebung(extend=True)
            sage: V(w)
            (0, x, y, z, t)

        TESTS::

            sage: R.<x,y> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=2)
            sage: V = W.verschiebung()
            sage: TestSuite(V).run(skip="_test_category")
            sage: V = W.verschiebung(extend=True)
            sage: TestSuite(V).run(skip="_test_category")
        """
        if extend:
            prec = domain.precision()

            if isinstance(domain, WittVectorRing_finotti):
                self._call_ = self._call_extend
                algorithm = "finotti"
            elif isinstance(domain, WittVectorRing_phantom):
                self._call_ = self._call_phantom_extend
                algorithm = "phantom"
            elif isinstance(domain, WittVectorRing_pinvertible):
                self._call_ = self._call_extend
                algorithm = "p_invertible"
            elif isinstance(domain, WittVectorRing_standard):
                self._call_ = self._call_extend
                algorithm = "standard"

            codomain = WittVectorRing(domain.coefficient_ring(), prec=prec+1,
                                      p=domain.prime(), algorithm=algorithm)

        else:
            if isinstance(domain, WittVectorRing_phantom):
                self._call_ = self._call_phantom_no_extend
            else:
                self._call_ = self._call_no_extend

            codomain = domain

        super().__init__(domain.Hom(codomain, CommutativeAdditiveGroups()))

    def _call_extend(self, x):
        """
        Evaluate ``self`` at ``x`` for the extended Verschiebung.

        EXAMPLES::

            sage: W = WittVectorRing(ZZ, p=11, prec=2)
            sage: w = W([1, 100])
            sage: V = W.verschiebung(extend=True)
            sage: V(w)  # indirect doctest
            (0, 1, 100)

        TESTS::

            sage: TestSuite(V).run(skip="_test_category")
        """
        Wcod = self.codomain()
        R = Wcod.coefficient_ring()
        return Wcod((R.zero(),) + x.coordinates())

    def _call_no_extend(self, x):
        """
        Evaluate ``self`` at ``x`` when the domain is the codomain.

        EXAMPLES::

            sage: W = WittVectorRing(GF(7), prec=4, algorithm="finotti")
            sage: w = W([1, 2, 3, 4])
            sage: V = W.verschiebung()
            sage: V(w)  # indirect doctest
            (0, 1, 2, 3)

        TESTS::

            sage: TestSuite(V).run(skip="_test_category")
        """
        W = self.domain()
        R = W.coefficient_ring()
        return W((R.zero(),) + x.coordinates())

    def _call_phantom_extend(self, x):
        """
        Evaluate ``self`` at ``x`` for the extended Verschiebung.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=3, algorithm="phantom")
            sage: w = W([x, x + y, x*y + y])
            sage: V = W.verschiebung(extend=True)
            sage: V(w)  # indirect doctest
            (0, x, x + y, x*y + y)

        TESTS::

            sage: TestSuite(V).run(skip="_test_category")
        """
        Wcod = self.codomain()
        p = Wcod.prime()
        phantom = x.phantom(lift=True)
        return Wcod(phantom=((phantom[0].parent().zero(),)
                             + tuple(p*c for c in phantom)))

    def _call_phantom_no_extend(self, x):
        """
        Evaluate ``self`` at ``x`` when the domain is the codomain.

        EXAMPLES::

            sage: R.<x,y> = PolynomialRing(GF(5))
            sage: W = WittVectorRing(R, prec=3, algorithm="phantom")
            sage: w = W([x, x + y, x*y + y])
            sage: V = W.verschiebung()
            sage: V(w)  # indirect doctest
            (0, x, x + y)

        TESTS::

            sage: TestSuite(V).run(skip="_test_category")
        """
        W = self.domain()
        p = W.prime()
        phantom = x.phantom(lift=True)
        return W(phantom=((phantom[0].parent().zero(),)
                          + tuple(p*c for c in phantom[:-1])))

    def _latex_(self) -> str:
        r"""
        Return a `\LaTeX` representation of ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(PolynomialRing(GF(3),'t'))
            sage: latex(W.verschiebung())
            V
        """
        return "V"

    def _repr_(self) -> str:
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: W = WittVectorRing(GF(11), prec=5)
            sage: W.verschiebung()
            Verschiebung map on the Ring of truncated 11-typical Witt vectors
            of length 5 over Finite Field of size 11
            sage: W.verschiebung(extend=True)
            Verschiebung map from the Ring of truncated 11-typical Witt vectors
            of length 5 over Finite Field of size 11 to the Ring of truncated
            11-typical Witt vectors of length 6 over Finite Field of size 11
        """
        if self.domain() is self.codomain():
            return f"Verschiebung map on the {self.domain()}"
        return f"Verschiebung map from the {self.domain()} to the "\
               f"{self.codomain()}"



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_blas.h":

  # Level 1
  int  gsl_blas_ddot(gsl_vector * x, gsl_vector * y, double * result)

  int  gsl_blas_zdotu(gsl_vector_complex * x, gsl_vector_complex * y, gsl_complex * dotu)

  int  gsl_blas_zdotc(gsl_vector_complex * x, gsl_vector_complex * y, gsl_complex * dotc)

  double  gsl_blas_dnrm2(gsl_vector * x)

  double  gsl_blas_dasum(gsl_vector * x)

  double  gsl_blas_dznrm2(gsl_vector_complex * x)

  double  gsl_blas_dzasum(gsl_vector_complex * x)

  CBLAS_INDEX_t  gsl_blas_idamax(gsl_vector * x)

  CBLAS_INDEX_t  gsl_blas_izamax(gsl_vector_complex * x)

  int  gsl_blas_dswap(gsl_vector * x, gsl_vector * y)

  int  gsl_blas_dcopy(gsl_vector * x, gsl_vector * y)

  int  gsl_blas_daxpy(double alpha, gsl_vector * x, gsl_vector * y)

  int  gsl_blas_zswap(gsl_vector_complex * x, gsl_vector_complex * y)

  int  gsl_blas_zcopy(gsl_vector_complex * x, gsl_vector_complex * y)

  int  gsl_blas_zaxpy(gsl_complex alpha, gsl_vector_complex * x, gsl_vector_complex * y)

  int  gsl_blas_drotg(double a[], double b[], double c[], double s[])

  int  gsl_blas_drotmg(double d1[], double d2[], double b1[], double b2, double P[])

  int  gsl_blas_drot(gsl_vector * x, gsl_vector * y, double c, double s)

  int  gsl_blas_drotm(gsl_vector * x, gsl_vector * y, double P[])

  void  gsl_blas_dscal(double alpha, gsl_vector * x)

  void  gsl_blas_zscal(gsl_complex alpha, gsl_vector_complex * x)

  void  gsl_blas_zdscal(double alpha, gsl_vector_complex * x)


  # Level 2
  int  gsl_blas_dgemv(CBLAS_TRANSPOSE_t TransA, double alpha, gsl_matrix * A, gsl_vector * x, double beta, gsl_vector * y)

  int  gsl_blas_dtrmv(CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t TransA, CBLAS_DIAG_t Diag, gsl_matrix * A, gsl_vector * x)

  int  gsl_blas_dtrsv(CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t TransA, CBLAS_DIAG_t Diag, gsl_matrix * A, gsl_vector * x)

  int  gsl_blas_zgemv (CBLAS_TRANSPOSE_t TransA,  gsl_complex alpha,  gsl_matrix_complex * A,  gsl_vector_complex * X,  gsl_complex beta, gsl_vector_complex * Y)

  int  gsl_blas_ztrmv (CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t TransA, CBLAS_DIAG_t Diag,  gsl_matrix_complex * A, gsl_vector_complex * X)

  int  gsl_blas_ztrsv (CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t TransA, CBLAS_DIAG_t Diag,  gsl_matrix_complex * A, gsl_vector_complex *X)

  int  gsl_blas_dsymv(CBLAS_UPLO_t Uplo, double alpha, gsl_matrix * A, gsl_vector * x, double beta, gsl_vector * y)

  int  gsl_blas_dger(double alpha, gsl_vector * x, gsl_vector * y, gsl_matrix * A)

  int  gsl_blas_dsyr(CBLAS_UPLO_t Uplo, double alpha, gsl_vector * x, gsl_matrix * A)

  int  gsl_blas_dsyr2(CBLAS_UPLO_t Uplo, double alpha, gsl_vector * x, gsl_vector * y, gsl_matrix * A)

  int  gsl_blas_zhemv (CBLAS_UPLO_t Uplo,  gsl_complex alpha,  gsl_matrix_complex * A,  gsl_vector_complex * X,  gsl_complex beta, gsl_vector_complex * Y)

  int  gsl_blas_zgeru ( gsl_complex alpha,  gsl_vector_complex * X,  gsl_vector_complex * Y, gsl_matrix_complex * A)

  int  gsl_blas_zgerc ( gsl_complex alpha,  gsl_vector_complex * X,  gsl_vector_complex * Y, gsl_matrix_complex * A)

  int  gsl_blas_zher (CBLAS_UPLO_t Uplo, double alpha,  gsl_vector_complex * X, gsl_matrix_complex * A)

  int  gsl_blas_zher2 (CBLAS_UPLO_t Uplo,  gsl_complex alpha,  gsl_vector_complex * X,  gsl_vector_complex * Y, gsl_matrix_complex * A)


  # Level 3
  int  gsl_blas_dgemm(CBLAS_TRANSPOSE_t TransA, CBLAS_TRANSPOSE_t TransB, double alpha, gsl_matrix * A, gsl_matrix * B, double beta, gsl_matrix * C)

  int  gsl_blas_dsymm(CBLAS_SIDE_t Side, CBLAS_UPLO_t Uplo, double alpha, gsl_matrix * A, gsl_matrix * B, double beta, gsl_matrix * C)

  int  gsl_blas_dsyrk(CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t Trans, double alpha, gsl_matrix * A, double beta, gsl_matrix * C)

  int  gsl_blas_dsyr2k(CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t Trans, double alpha, gsl_matrix * A, gsl_matrix * B, double beta, gsl_matrix * C)

  int  gsl_blas_dtrmm(CBLAS_SIDE_t Side, CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t TransA, CBLAS_DIAG_t Diag, double alpha, gsl_matrix * A, gsl_matrix * B)

  int  gsl_blas_dtrsm(CBLAS_SIDE_t Side, CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t TransA, CBLAS_DIAG_t Diag, double alpha, gsl_matrix * A, gsl_matrix * B)

  int  gsl_blas_zgemm (CBLAS_TRANSPOSE_t TransA, CBLAS_TRANSPOSE_t TransB,  gsl_complex alpha,  gsl_matrix_complex * A,  gsl_matrix_complex * B,  gsl_complex beta, gsl_matrix_complex * C)

  int  gsl_blas_zsymm (CBLAS_SIDE_t Side, CBLAS_UPLO_t Uplo,  gsl_complex alpha,  gsl_matrix_complex * A,  gsl_matrix_complex * B,  gsl_complex beta, gsl_matrix_complex * C)

  int  gsl_blas_zsyrk (CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t Trans,  gsl_complex alpha,  gsl_matrix_complex * A,  gsl_complex beta, gsl_matrix_complex * C)

  int  gsl_blas_zsyr2k (CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t Trans,  gsl_complex alpha,  gsl_matrix_complex * A,  gsl_matrix_complex * B,  gsl_complex beta, gsl_matrix_complex *C)

  int  gsl_blas_ztrmm (CBLAS_SIDE_t Side, CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t TransA, CBLAS_DIAG_t Diag,  gsl_complex alpha,  gsl_matrix_complex * A, gsl_matrix_complex * B)

  int  gsl_blas_ztrsm (CBLAS_SIDE_t Side, CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t TransA, CBLAS_DIAG_t Diag,  gsl_complex alpha,  gsl_matrix_complex * A, gsl_matrix_complex * B)

  int  gsl_blas_zhemm (CBLAS_SIDE_t Side, CBLAS_UPLO_t Uplo,  gsl_complex alpha,  gsl_matrix_complex * A,  gsl_matrix_complex * B,  gsl_complex beta, gsl_matrix_complex * C)

  int  gsl_blas_zherk (CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t Trans, double alpha,  gsl_matrix_complex * A, double beta, gsl_matrix_complex * C)

  int  gsl_blas_zher2k (CBLAS_UPLO_t Uplo, CBLAS_TRANSPOSE_t Trans,  gsl_complex alpha,  gsl_matrix_complex * A,  gsl_matrix_complex * B, double beta, gsl_matrix_complex * C)



"""
Tests for the Sage/Pynac interaction
"""

# ****************************************************************************
#       Copyright (C) 2013 Volker Braun <vbraun.name@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


def rational_powers_memleak():
    """
    Check that there is no memory leak in rational powers.

    OUTPUT: boolean; whether the memory leak was detected

    See :issue:`9129`.

    EXAMPLES::

        sage: from sage.symbolic.tests import rational_powers_memleak
        sage: rational_powers_memleak()
        False
    """
    from sage.rings.integer_ring import ZZ
    import gc
    gc.collect()
    c0 = sum(1 for obj in gc.get_objects())
    for i in range(1000):
        _ = ZZ(2).sqrt()
    gc.collect()
    c1 = sum(1 for obj in gc.get_objects())
    # Test that we do not leak an object at each iteration
    return (c1 - c0) >= 1000



# sage.doctest: needs sage.libs.flint
"""
Elements of quasimodular forms rings

AUTHORS:

- DAVID AYOTTE (2021-03-18): initial version
- Seewoo Lee (2023-09): coefficients method
"""
# ****************************************************************************
#       Copyright (C) 2021 David Ayotte
#                     2023 Seewoo Lee <seewoo5@berkeley.edu>
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from typing import Self

from sage.modular.arithgroup.congroup_sl2z import SL2Z_class
from sage.modular.modform.constructor import EisensteinForms
from sage.modular.modform.eis_series import eisenstein_series_qexp
from sage.modular.modform.element import GradedModularFormElement

from sage.structure.element import ModuleElement
from sage.structure.richcmp import richcmp, op_NE, op_EQ

from sage.rings.integer import Integer
from sage.rings.polynomial.polynomial_element import Polynomial
from sage.rings.integer_ring import ZZ


class QuasiModularFormsElement(ModuleElement):
    r"""
    A quasimodular forms ring element. Such an element is described by
    SageMath as a polynomial

    .. MATH::

        F = f_0 + f_1 E_2 + f_2 E_2^2 + \cdots + f_m E_2^m

    where each `f_i` a graded modular form element
    (see :class:`~sage.modular.modform.element.GradedModularFormElement`)

    For an integer `k`, we say that `F` is homogeneous of weight `k` if
    it lies in an homogeneous component of degree `k` of the graded ring
    of quasimodular forms.

    EXAMPLES::

        sage: QM = QuasiModularForms(1)
        sage: QM.gens()
        (1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6),
        1 + 240*q + 2160*q^2 + 6720*q^3 + 17520*q^4 + 30240*q^5 + O(q^6),
        1 - 504*q - 16632*q^2 - 122976*q^3 - 532728*q^4 - 1575504*q^5 + O(q^6))
        sage: QM.0 + QM.1
        2 + 216*q + 2088*q^2 + 6624*q^3 + 17352*q^4 + 30096*q^5 + O(q^6)
        sage: QM.0 * QM.1
        1 + 216*q - 3672*q^2 - 62496*q^3 - 322488*q^4 - 1121904*q^5 + O(q^6)
        sage: (QM.0)^2
        1 - 48*q + 432*q^2 + 3264*q^3 + 9456*q^4 + 21600*q^5 + O(q^6)
        sage: QM.0 == QM.1
        False

    Quasimodular forms ring element can be created via a polynomial in `E2` over the ring of modular forms::

        sage: E2 = QM.polygen()
        sage: E2.parent()
        Univariate Polynomial Ring in E2 over Ring of Modular Forms for Modular Group SL(2,Z) over Rational Field
        sage: QM(E2)
        1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6)
        sage: M = QM.modular_forms_subring()
        sage: QM(M.0 * E2 + M.1 * E2^2)
        2 - 336*q + 4320*q^2 + 398400*q^3 - 3772992*q^4 - 89283168*q^5 + O(q^6)

    One may convert a quasimodular form into a multivariate polynomial in the
    generators of the ring by calling
    :meth:`~sage.modular.quasimodform.element.QuasiModularFormsElement.polynomial`::

        sage: QM = QuasiModularForms(1)
        sage: F = QM.0^2 + QM.1^2 + QM.0*QM.1*QM.2
        sage: F.polynomial()
        E2*E4*E6 + E4^2 + E2^2

    If the group is not the full modular group, the default names of the
    generators are given by ``Ek_i`` and ``Sk_i`` to denote the `i`-th basis
    element of the weight `k` Eisenstein subspace and cuspidal subspace
    respectively (for more details, see the documentation of
    :meth:`~sage.modular.quasimodform.ring.QuasiModularFormsRing.polynomial_ring`) ::

        sage: QM = QuasiModularForms(Gamma1(4))
        sage: F = (QM.0^4)*(QM.1^3) + QM.3
        sage: F.polynomial()
        -512*E2^4*E2_1^3 + E2^4*E3_0^2 + 48*E2^4*E3_1^2 + E3_0
    """
    def __init__(self, parent, polynomial) -> None:
        r"""
        INPUT:

        - ``parent`` -- a quasimodular forms ring
        - ``polynomial`` -- a polynomial `f_0 + f_1 E_2 + ... + f_n E_2^n` where
          each `f_i` are modular forms ring elements and `E_2` correspond to the
          weight 2 Eisenstein series

        OUTPUT: ``QuasiModularFormsElement``

        TESTS::

            sage: QM = QuasiModularForms(1)
            sage: QM.element_class(QM, 'E2')
            Traceback (most recent call last):
            ...
            TypeError: 'polynomial' argument should be of type 'Polynomial'
            sage: x = polygen(QQ)
            sage: QM.element_class(QM, x^2 + 1)
            Traceback (most recent call last):
            ...
            ValueError: at least one coefficient is not a 'GradedModularFormElement'
        """
        if not isinstance(polynomial, Polynomial):
            raise TypeError("'polynomial' argument should be of type 'Polynomial'")
        for f in polynomial.coefficients():
            if not isinstance(f, GradedModularFormElement):
                raise ValueError("at least one coefficient is not a 'GradedModularFormElement'")
        self._polynomial = polynomial
        ModuleElement.__init__(self, parent)

    def q_expansion(self, prec=6):
        r"""
        Return the `q`-expansion of the given quasimodular form up to precision
        ``prec`` (default: 6).

        An alias of this method is ``qexp``.

        EXAMPLES::

            sage: QM = QuasiModularForms()
            sage: E2 = QM.0
            sage: E2.q_expansion()
            1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6)
            sage: E2.q_expansion(prec=10)
            1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 - 288*q^6 - 192*q^7 - 360*q^8 - 312*q^9 + O(q^10)
        """
        E2 = eisenstein_series_qexp(2, prec=prec, K=self.base_ring(), normalization='constant')  # normalization -> to force integer coefficients
        coefficients = self._polynomial.coefficients(sparse=False)
        return sum(f.q_expansion(prec=prec) * E2**idx for idx, f in enumerate(coefficients))

    qexp = q_expansion  # alias

    def _repr_(self) -> str:
        r"""
        String representation of ``self``.

        TESTS::

            sage: QM = QuasiModularForms()
            sage: QM.0
            1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6)
            sage: QM.1
            1 + 240*q + 2160*q^2 + 6720*q^3 + 17520*q^4 + 30240*q^5 + O(q^6)
            sage: QM.2
            1 - 504*q - 16632*q^2 - 122976*q^3 - 532728*q^4 - 1575504*q^5 + O(q^6)
        """
        return str(self.q_expansion())

    def _latex_(self) -> str:
        r"""
        Return a latex representation of ``self``.

        TESTS::

            sage: QM = QuasiModularForms(1)
            sage: latex(QM.0)
            1 - 24 q - 72 q^{2} - 96 q^{3} - 168 q^{4} - 144 q^{5} + O(q^{6})
        """
        return self.q_expansion()._latex_()

    def _richcmp_(self, other, op) -> bool:
        r"""
        Compare ``self`` with ``other``.

        TESTS::

            sage: QM = QuasiModularForms(1)
            sage: QM.0 == QM.1
            False
            sage: QM.0 == QM.0
            True
            sage: QM.0 != QM.1
            True
            sage: QM.0 != QM.0
            False
            sage: QM.0 < QM.1
            Traceback (most recent call last):
            ...
            TypeError: invalid comparison between quasimodular forms ring elements
        """
        if op != op_EQ and op != op_NE:
            raise TypeError('invalid comparison between quasimodular forms ring elements')
        return richcmp(self._polynomial, other._polynomial, op)

    def _add_(self, other) -> Self:
        r"""
        Addition of two ``QuasiModularFormElement``.

        INPUT:

        - ``other`` -- ``QuasiModularFormElement``

        OUTPUT: a ``QuasiModularFormElement``

        TESTS::

            sage: QM = QuasiModularForms(1)
            sage: QM.0 + QM.1
            2 + 216*q + 2088*q^2 + 6624*q^3 + 17352*q^4 + 30096*q^5 + O(q^6)
            sage: QM.0 + (QM.1 + QM.2) == (QM.0 + QM.1) + QM.2
            True
            sage: QM = QuasiModularForms(5)
            sage: QM.0 + QM.1 + QM.2 + QM.3
            3 - 17*q - 54*q^2 - 62*q^3 - 98*q^4 + 137*q^5 + O(q^6)
        """
        return self.__class__(self.parent(), self._polynomial + other._polynomial)

    def __neg__(self) -> Self:
        r"""
        The negation of ``self``.

        TESTS::

            sage: -QuasiModularForms(1).0
            -1 + 24*q + 72*q^2 + 96*q^3 + 168*q^4 + 144*q^5 + O(q^6)
            sage: QuasiModularForms(1).0 - QuasiModularForms(1).0
            0
            sage: -QuasiModularForms(Gamma1(2)).2
            -1 - 240*q^2 - 2160*q^4 + O(q^6)
        """
        return self.__class__(self.parent(), -self._polynomial)

    def _mul_(self, other) -> Self:
        r"""
        The multiplication of two ``QuasiModularFormElement``.

        INPUT:

        - ``other`` -- ``QuasiModularFormElement``

        OUTPUT: a ``QuasiModularFormElement``

        TESTS::

            sage: QM = QuasiModularForms(1)
            sage: QM.0 * QM.1
            1 + 216*q - 3672*q^2 - 62496*q^3 - 322488*q^4 - 1121904*q^5 + O(q^6)
            sage: (QM.0 * QM.1) * QM.2 == QM.0 * (QM.1 * QM.2)
            True
            sage: QM = QuasiModularForms(Gamma1(5))
            sage: QM.0 * QM.1 * QM.2
            q - 24*q^2 - 66*q^3 - 189*q^4 - 1917*q^5 + O(q^6)
        """
        return self.__class__(self.parent(), self._polynomial * other._polynomial)

    def _lmul_(self, c) -> Self:
        r"""
        The left action of the base ring on ``self``.

        INPUT:

        - ``other`` -- ``QuasiModularFormElement``

        OUTPUT: a ``QuasiModularFormElement``

        TESTS::

            sage: QM = QuasiModularForms(1)
            sage: (1/2) * QM.0
            1/2 - 12*q - 36*q^2 - 48*q^3 - 84*q^4 - 72*q^5 + O(q^6)
            sage: QM.0 * (3/2)
            3/2 - 36*q - 108*q^2 - 144*q^3 - 252*q^4 - 216*q^5 + O(q^6)
            sage: (5/2) * QuasiModularForms(Gamma0(7)).0 * (3/2)
            15/4 - 90*q - 270*q^2 - 360*q^3 - 630*q^4 - 540*q^5 + O(q^6)
        """
        return self.__class__(self.parent(), c * self._polynomial)

    def __bool__(self) -> bool:
        r"""
        Return whether ``self`` is nonzero.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: bool(QM(0))
            False
            sage: bool(QM(1))
            True
            sage: bool(QM.0)
            True
        """
        return bool(self._polynomial)

    def depth(self):
        r"""
        Return the depth of this quasimodular form.

        Note that the quasimodular form must be homogeneous of weight
        `k`. Recall that the *depth* is the integer `p` such that

        .. MATH::

            f = f_0 + f_1 E_2 + \cdots + f_p E_2^p,

        where `f_i` is a modular form of weight `k - 2i` and `f_p` is
        nonzero.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: E2, E4, E6 = QM.gens()
            sage: E2.depth()
            1
            sage: F = E4^2 + E6*E2 + E4*E2^2 + E2^4
            sage: F.depth()
            4
            sage: QM(7/11).depth()
            0

        TESTS::

            sage: QM = QuasiModularForms(1)
            sage: (QM.0 + QM.1).depth()
            Traceback (most recent call last):
            ...
            ValueError: the given graded quasiform is not an homogeneous element
        """
        if not self.is_homogeneous():
            raise ValueError("the given graded quasiform is not an "
                             "homogeneous element")
        return self._polynomial.degree()

    def is_zero(self) -> bool:
        r"""
        Return whether the given quasimodular form is zero.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: QM.zero().is_zero()
            True
            sage: QM(0).is_zero()
            True
            sage: QM(1/2).is_zero()
            False
            sage: (QM.0).is_zero()
            False
            sage: QM = QuasiModularForms(Gamma0(2))
            sage: QM(0).is_zero()
            True
        """
        return not self

    def is_one(self) -> bool:
        r"""
        Return whether the given quasimodular form is 1, i.e. the
        multiplicative identity.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: QM.one().is_one()
            True
            sage: QM(1).is_one()
            True
            sage: (QM.0).is_one()
            False
            sage: QM = QuasiModularForms(Gamma0(2))
            sage: QM(1).is_one()
            True
        """
        return self._polynomial.is_one()

    def is_graded_modular_form(self) -> bool:
        r"""
        Return whether the given quasimodular form is a
        graded modular form element
        (see :class:`~sage.modular.modform.element.GradedModularFormElement`).

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: (QM.0).is_graded_modular_form()
            False
            sage: (QM.1).is_graded_modular_form()
            True
            sage: (QM.1 + QM.0^2).is_graded_modular_form()
            False
            sage: (QM.1^2 + QM.2).is_graded_modular_form()
            True
            sage: QM = QuasiModularForms(Gamma0(6))
            sage: (QM.0).is_graded_modular_form()
            False
            sage: (QM.1 + QM.2 + QM.1 * QM.3).is_graded_modular_form()
            True
            sage: QM.zero().is_graded_modular_form()
            True
            sage: QM = QuasiModularForms(Gamma0(6))
            sage: (QM.0).is_graded_modular_form()
            False
            sage: (QM.0 + QM.1*QM.2 + QM.3).is_graded_modular_form()
            False
            sage: (QM.1*QM.2 + QM.3).is_graded_modular_form()
            True

        .. NOTE::

            A graded modular form in SageMath is not necessarily a modular form
            as it can have mixed weight components. To check for modular forms
            only, see the method :meth:`is_modular_form`.
        """
        return self._polynomial.degree() <= 0

    def is_modular_form(self) -> bool:
        r"""
        Return whether the given quasimodular form is a modular form.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: (QM.0).is_modular_form()
            False
            sage: (QM.1).is_modular_form()
            True
            sage: (QM.1 + QM.2).is_modular_form() # mixed weight components
            False
            sage: QM.zero().is_modular_form()
            True
            sage: QM = QuasiModularForms(Gamma0(4))
            sage: (QM.0).is_modular_form()
            False
            sage: (QM.1).is_modular_form()
            True
        """
        return self._polynomial.degree() <= 0 and self._polynomial[0].is_modular_form()

    def polynomial(self, names=None):
        r"""
        Return a multivariate polynomial such that every variable corresponds to
        a generator of the ring, ordered by the method:
        :meth:`~sage.modular.quasimodform.ring.QuasiModularForms.gens`.

        An alias of this method is ``to_polynomial``.

        INPUT:

        - ``names``-- string (default: ``None``); list or tuple of names
          (strings), or a comma separated string. Defines the names for the
          generators of the multivariate polynomial ring. The default names are
          of the form ``ABCk`` where ``k`` is a number corresponding to the
          weight of the form ``ABC``.

        OUTPUT: a multivariate polynomial in the variables ``names``

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: (QM.0 + QM.1).polynomial()
            E4 + E2
            sage: (1/2 + QM.0 + 2*QM.1^2 + QM.0*QM.2).polynomial()
            E2*E6 + 2*E4^2 + E2 + 1/2

        Check that :issue:`34569` is fixed::

            sage: QM = QuasiModularForms(Gamma1(3))
            sage: QM.ngens()
            5
            sage: (QM.0 + QM.1 + QM.2*QM.1 + QM.3*QM.4).polynomial()
            E3_1*E4_0 + E2_0*E3_0 + E2 + E2_0
        """
        P = self.parent().polynomial_ring(names)
        poly_gens = P.gens()
        E2 = poly_gens[0]
        poly_gens = poly_gens[1:]
        modform_poly_gens = self.parent().modular_forms_subring().polynomial_ring(names='x').gens()
        subs_dictionary = {}
        for idx, g in enumerate(modform_poly_gens):
            subs_dictionary[g] = poly_gens[idx]
        return sum(f.to_polynomial().subs(subs_dictionary) * E2 ** exp for exp, f in enumerate(self._polynomial.coefficients(sparse=False)))

    to_polynomial = polynomial  # alias

    def weights_list(self) -> list[Integer]:
        r"""
        Return the list of the weights of all the graded components of the given
        graded quasimodular form.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: (QM.0).weights_list()
            [2]
            sage: (QM.0 + QM.1 + QM.2).weights_list()
            [2, 4, 6]
            sage: (QM.0 * QM.1 + QM.2).weights_list()
            [6]
            sage: QM(1/2).weights_list()
            [0]
            sage: QM = QuasiModularForms(Gamma1(3))
            sage: (QM.0 + QM.1 + QM.2*QM.1 + QM.3*QM.4).weights_list()
            [2, 5, 7]
        """
        return sorted(self.homogeneous_components().keys())

    def is_homogeneous(self) -> bool:
        r"""
        Return whether the graded quasimodular form is a homogeneous element,
        that is, it lives in a unique graded components of the parent of
        ``self``.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: E2, E4, E6 = QM.gens()
            sage: (E2).is_homogeneous()
            True
            sage: (E2 + E4).is_homogeneous()
            False
            sage: (E2 * E4 + E6).is_homogeneous()
            True
            sage: QM(1).is_homogeneous()
            True
            sage: (1 + E2).is_homogeneous()
            False
            sage: F = E6^3 + E4^4*E2 + (E4^2*E6)*E2^2 + (E4^3 + E6^2)*E2^3
            sage: F.is_homogeneous()
            True
        """
        k = None
        for i, c in enumerate(self._polynomial.coefficients(sparse=False)):
            if c:
                if not c.is_homogeneous():
                    return False
                if k is None:
                    k = c.weight() + 2*i
                    continue
                if c.weight() + 2*i != k:
                    return False
        return True

    def weight(self) -> Integer:
        r"""
        Return the weight of the given quasimodular form.

        Note that the given form must be homogeneous. An alias of this method is
        ``degree``.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: (QM.0).weight()
            2
            sage: (QM.0 * QM.1 + QM.2).weight()
            6
            sage: QM(1/2).weight()
            0
            sage: (QM.0).degree()
            2
            sage: (QM.0 + QM.1).weight()
            Traceback (most recent call last):
            ...
            ValueError: the given graded quasiform is not an homogeneous element
        """
        if self.is_homogeneous():
            return (self._polynomial.leading_coefficient().weight()
                    + 2*self._polynomial.degree())
        raise ValueError("the given graded quasiform is not an homogeneous "
                         "element")

    degree = weight  # alias

    def homogeneous_components(self) -> dict[Integer, Self]:
        r"""
        Return a dictionary where the values are the homogeneous components of
        the given graded form and the keys are the weights of those components.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: (QM.0).homogeneous_components()
            {2: 1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6)}
            sage: (QM.0 + QM.1 + QM.2).homogeneous_components()
            {2: 1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6),
             4: 1 + 240*q + 2160*q^2 + 6720*q^3 + 17520*q^4 + 30240*q^5 + O(q^6),
             6: 1 - 504*q - 16632*q^2 - 122976*q^3 - 532728*q^4 - 1575504*q^5 + O(q^6)}
            sage: (1 + QM.0).homogeneous_components()
            {0: 1, 2: 1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6)}
            sage: QM5 = QuasiModularForms(Gamma1(3))
            sage: F = QM.1 + QM.1*QM.2 + QM.1*QM.0 + (QM.1 + QM.2^2)*QM.0^3
            sage: F.homogeneous_components()
            {4: 1 + 240*q + 2160*q^2 + 6720*q^3 + 17520*q^4 + 30240*q^5 + O(q^6),
             6: 1 + 216*q - 3672*q^2 - 62496*q^3 - 322488*q^4 - 1121904*q^5 + O(q^6),
             10: 2 - 96*q - 149040*q^2 - 4986240*q^3 - 67535952*q^4 - 538187328*q^5 + O(q^6),
             18: 1 - 1080*q + 294840*q^2 - 902880*q^3 - 452402280*q^4 + 105456816*q^5 + O(q^6)}
            sage: F = QM.zero()
            sage: F.homogeneous_components()
            {0: 0}
            sage: F = QM(42/13)
            sage: F.homogeneous_components()
            {0: 42/13}
        """
        QM = self.parent()
        if self.is_zero():
            return {ZZ.zero(): self}
        components: dict[Integer, Self] = {}
        E2 = self.parent().weight_2_eisenstein_series()
        for i, c in enumerate(self._polynomial.coefficients(sparse=False)):
            if c:
                forms = c._forms_dictionary
                for k in forms:
                    try:
                        components[ZZ(k + 2*i)] += QM(forms[k]*(E2**i))
                    except KeyError:
                        components[ZZ(k + 2*i)] = QM(forms[k]*(E2**i))
        return components

    def __getitem__(self, weight) -> Self | None:
        r"""
        Return the homogeneous component of the given quasimodular form ring
        element.

        An alias of this method is ``homogeneous_component``.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: E2, E4, E6 = QM.gens()
            sage: F = E2 + E4*E6 + E2^3*E6
            sage: F[2]
            1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6)
            sage: F[10]
            1 - 264*q - 135432*q^2 - 5196576*q^3 - 69341448*q^4 - 515625264*q^5 + O(q^6)
            sage: F[12]
            1 - 576*q + 21168*q^2 + 308736*q^3 - 15034608*q^4 - 39208320*q^5 + O(q^6)
            sage: F[4]
            0
            sage: F.homogeneous_component(2)
            1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 + O(q^6)

        TESTS::

            sage: F[x]                                                                  # needs sage.symbolic
            Traceback (most recent call last):
            ...
            KeyError: 'the weight must be an integer'
            sage: F[-1]
            Traceback (most recent call last):
            ...
            ValueError: the weight must be nonnegative
        """
        if not isinstance(weight, (int, Integer)):
            raise KeyError("the weight must be an integer")
        if weight < 0:
            raise ValueError("the weight must be nonnegative")
        return self.homogeneous_components().get(Integer(weight),
                                                 self.parent().zero())

    homogeneous_component = __getitem__  # alias

    def serre_derivative(self) -> Self:
        r"""
        Return the Serre derivative of the given quasimodular form.

        If the form is not homogeneous, then this method sums the Serre
        derivative of each homogeneous component.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: E2, E4, E6 = QM.gens()
            sage: DE2 = E2.serre_derivative(); DE2
            -1/6 - 16*q - 216*q^2 - 832*q^3 - 2248*q^4 - 4320*q^5 + O(q^6)
            sage: DE2 == (-E2^2 - E4)/12
            True
            sage: DE4 = E4.serre_derivative(); DE4
            -1/3 + 168*q + 5544*q^2 + 40992*q^3 + 177576*q^4 + 525168*q^5 + O(q^6)
            sage: DE4 == (-1/3) * E6
            True
            sage: DE6 = E6.serre_derivative(); DE6
            -1/2 - 240*q - 30960*q^2 - 525120*q^3 - 3963120*q^4 - 18750240*q^5 + O(q^6)
            sage: DE6 == (-1/2) * E4^2
            True

        The Serre derivative raises the weight of homogeneous elements by 2::

            sage: F = E6 + E4 * E2
            sage: F.weight()
            6
            sage: F.serre_derivative().weight()
            8

        Check that :issue:`34569` is fixed::

            sage: QM = QuasiModularForms(Gamma1(3))
            sage: E2 = QM.weight_2_eisenstein_series()
            sage: E2.serre_derivative()
            -1/6 - 16*q - 216*q^2 - 832*q^3 - 2248*q^4 - 4320*q^5 + O(q^6)
            sage: F = QM.0 + QM.1*QM.2
        """
        # initial variables:
        QM = self.parent()
        R = QM.base_ring()
        E2 = QM.gen(0)
        if isinstance(QM.group(), SL2Z_class):
            E4 = QM.gen(1)
        else:
            E4 = QM(EisensteinForms(group=1, weight=4, base_ring=R).gen(0))

        # compute the derivative of E2: q*dE2/dq
        E2deriv = R(12).inverse_of_unit() * (E2 ** 2 - E4)

        # sum the Serre derivative of each monomial of the form: f * E2^n
        # they are equal to:
        # [E2^n * serre_deriv(f)]  +  [n * f * E2^(n-1) * D(E2)]  -  [n/6 * f * E2^(n+1)]
        #   =      A               +              B               -           C
        der = QM.zero()
        u6 = R(6).inverse_of_unit()
        for n, f in enumerate(self._polynomial.coefficients(sparse=False)):
            if n == 0:
                der += QM(f.serre_derivative())
            else:
                A = (E2 ** n) * f.serre_derivative()
                B = R(n) * f * E2 ** (n - 1) * E2deriv
                C = R(n) * u6 * E2 ** (n + 1) * f
                der += QM(A + B - C)
        return der

    def derivative(self) -> Self:
        r"""
        Return the derivative `q \frac{d}{dq}` of the given quasimodular form.

        If the form is not homogeneous, then this method sums the derivative of
        each homogeneous component.

        EXAMPLES::

            sage: QM = QuasiModularForms(1)
            sage: E2, E4, E6 = QM.gens()
            sage: dE2 = E2.derivative(); dE2
            -24*q - 144*q^2 - 288*q^3 - 672*q^4 - 720*q^5 + O(q^6)
            sage: dE2 == (E2^2 - E4)/12 # Ramanujan identity
            True
            sage: dE4 = E4.derivative(); dE4
            240*q + 4320*q^2 + 20160*q^3 + 70080*q^4 + 151200*q^5 + O(q^6)
            sage: dE4 == (E2 * E4 - E6)/3 # Ramanujan identity
            True
            sage: dE6 = E6.derivative(); dE6
            -504*q - 33264*q^2 - 368928*q^3 - 2130912*q^4 - 7877520*q^5 + O(q^6)
            sage: dE6 == (E2 * E6 - E4^2)/2 # Ramanujan identity
            True

        Note that the derivative of a modular form is not necessarily a modular form::

            sage: dE4.is_modular_form()
            False
            sage: dE4.weight()
            6
        """
        QM = self.parent()
        E2 = QM.gen(0)
        R = self.base_ring()
        u = R(12).inverse_of_unit()
        hom_comp = self.homogeneous_components()

        return sum(f.serre_derivative() + R(k) * u * f * E2 for k, f in hom_comp.items())

    def _compute(self, X: list) -> list:
        r"""
        Compute the coefficients of `q^n` of the `q`-expansion of this,
        graded quasimodular form for `n` in the list `X`.

        The results are not cached.  (Use coefficients for cached results).

        EXAMPLES::

            sage: E2 = QuasiModularForms(1).0
            sage: E2.q_expansion(10)
            1 - 24*q - 72*q^2 - 96*q^3 - 168*q^4 - 144*q^5 - 288*q^6 - 192*q^7 - 360*q^8 - 312*q^9 + O(q^10)
            sage: E2._compute([3, 6])
            [-96, -288]
            sage: E2._compute([])
            []
        """
        if not isinstance(X, list) or not X:
            return []
        bound = max(X)
        q_exp = self.q_expansion(bound + 1)
        return [q_exp[i] for i in X]

    def coefficients(self, X):
        r"""
        Return the coefficients of `q^n` of the `q`-expansion of this,
        graded quasimodular form for `n` in the list `X`.

        If X is an integer, return coefficients for indices from 1
        to X. This method caches the result.

        EXAMPLES::

            sage: E2, E4 = QuasiModularForms(1).0, QuasiModularForms(1).1
            sage: f = E2^2
            sage: g = E2^3 * E4
            sage: f.coefficients(10)
            [-48, 432, 3264, 9456, 21600, 39744, 66432, 105840, 147984, 220320]
            sage: f.coefficients([0,1])
            [1, -48]
            sage: f.coefficients([0,1,2,3])
            [1, -48, 432, 3264]
            sage: f.coefficients([2,3])
            [432, 3264]
            sage: g.coefficients(10)
            [168,
             -13608,
             210336,
             1805496,
             -22562064,
             -322437024,
             -2063087808,
             -9165872520,
             -32250917496,
             -96383477232]
            sage: g.coefficients([3, 7])
            [210336, -2063087808]
        """
        try:
            self.__coefficients
        except AttributeError:
            self.__coefficients = {}
        if isinstance(X, Integer):
            X = list(range(1, X + 1))
        Y = [n for n in X if n not in self.__coefficients]
        v = self._compute(Y)
        for i in range(len(v)):
            self.__coefficients[Y[i]] = v[i]
        return [self.__coefficients[x] for x in X]



r"""
Elements in Ore modules

AUTHOR:

- Xavier Caruso (2024-10)
"""

# ***************************************************************************
#    Copyright (C) 2024 Xavier Caruso <xavier.caruso@normalesup.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 2 of the License, or
#    (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.misc.repr import repr_lincomb
from sage.modules.free_module_element import FreeModuleElement_generic_dense


class OreModuleElement(FreeModuleElement_generic_dense):
    r"""
    A generic element of a Ore module.
    """
    def _repr_(self):
        r"""
        Return a string representation of this element.

        EXAMPLES::

            sage: K.<t> = Frac(QQ['t'])
            sage: S.<X> = OrePolynomialRing(K, K.derivation())
            sage: M.<v,w> = S.quotient_module((t+1)*X^2 + 1)

            sage: v - w  # indirect doctest
            v - w
            sage: w - v  # indirect doctest
            -v + w
            sage: X^5*v  # indirect doctest
            ((-4*t+2)/(t^4+4*t^3+6*t^2+4*t+1))*v + ((t-5)/(t^3+3*t^2+3*t+1))*w
        """
        parent = self.parent()
        names = parent._names
        if parent._names is None:
            return self.parent()._repr_element(self)
        return repr_lincomb([(names[i], self[i]) for i in range(len(names))])

    def _latex_(self):
        r"""
        Return a LaTeX representation of this element.

        EXAMPLES::

            sage: K.<t> = Frac(QQ['t'])
            sage: S.<X> = OrePolynomialRing(K, K.derivation())
            sage: M.<v,w> = S.quotient_module((t+1)*X^2 + 1)

            sage: latex(v - w)
            v - w
            sage: latex(w - v)
            -v + w
            sage: latex(X^5*v)
            \left(\frac{-4 t + 2}{t^{4} + 4 t^{3} + 6 t^{2} + 4 t + 1}\right) v + \left(\frac{t - 5}{t^{3} + 3 t^{2} + 3 t + 1}\right) w
        """
        parent = self.parent()
        if parent._names is None:
            return self.parent()._latex_element(self)
        names = parent._latex_names
        return repr_lincomb([(names[i], self[i]) for i in range(len(names))], is_latex=True)

    def is_mutable(self) -> bool:
        r"""
        Always return ``False`` since elements in Ore modules
        are all immutable.

        EXAMPLES::

            sage: K.<t> = Frac(QQ['t'])
            sage: S.<X> = OrePolynomialRing(K, K.derivation())
            sage: M = S.quotient_module(X^2 + t)

            sage: v, w = M.basis()
            sage: v
            (1, 0)
            sage: v.is_mutable()
            False
            sage: v[1] = 1
            Traceback (most recent call last):
            ...
            ValueError: vectors in Ore modules are immutable
        """
        return False

    def __setitem__(self, i, v):
        r"""
        Always raise an error since elements in Ore modules are
        all immutable.

        TESTS::

            sage: K.<t> = Frac(QQ['t'])
            sage: S.<X> = OrePolynomialRing(K, K.derivation())
            sage: M.<v,w> = S.quotient_module(X^2 + t)
            sage: w[1] = 0
            Traceback (most recent call last):
            ...
            ValueError: vectors in Ore modules are immutable
        """
        raise ValueError("vectors in Ore modules are immutable")

    def __hash__(self):
        r"""
        Return a hash of this element.

        TESTS::

            sage: K.<t> = Frac(QQ['t'])
            sage: S.<X> = OrePolynomialRing(K, K.derivation())
            sage: M.<v,w> = S.quotient_module(X^2 + t)
            sage: hash(v)  # random
            -5164621852614943976
            sage: hash(w)  # random
            -1950498447580522560
        """
        return hash(tuple(self))

    def vector(self):
        r"""
        Return the coordinates vector of this element.

        EXAMPLES::

            sage: K.<t> = Frac(QQ['t'])
            sage: S.<X> = OrePolynomialRing(K, K.derivation())
            sage: M.<v,w> = S.quotient_module(X^2 + t)
            sage: v.vector()
            (1, 0)

        We underline that this vector is not an element of the
        Ore module; it lives in `K^2`. Compare::

            sage: v.parent()
            Ore module <v, w> over Fraction Field of Univariate Polynomial Ring in t over Rational Field twisted by d/dt
            sage: v.vector().parent()
            Vector space of dimension 2 over Fraction Field of Univariate Polynomial Ring in t over Rational Field
        """
        V = self.parent().module()
        return V(self.list())

    def image(self, integral=False):
        r"""
        Return the image of this element by the pseudomorphism
        defining the action of the Ore variable on this Ore module.

        INPUT:

        - ``integral`` (default: ``False``) -- a boolean;
          if ``False``, allows for an output with coordinates in
          the fraction field of the base ring

        EXAMPLES::

            sage: A.<t> = QQ['t']
            sage: d = A.derivation()
            sage: S.<X> = OrePolynomialRing(A, A.derivation())
            sage: M.<v,w> = S.quotient_module(X^2 + t)
            sage: v.image()
            w
            sage: w.image()
            -t*v

        TESTS:

        We check that this corresponds to the action of `X`::

            sage: x = M.random_element()
            sage: x.image() == X*x
            True

        We check the behavior when the underlying Ore module has a nontrivial
        denominator. In this case, the output is an element of the scalar
        extension of the Ore module to its fraction field::

            sage: from sage.modules.ore_module import OreModule
            sage: mat = matrix(A, [[1, t], [0, t-1]])
            sage: N.<e0, e1> = OreModule(mat, d, t-1)
            sage: N.matrix()
            [1/(t - 1) t/(t - 1)]
            [        0         1]
            sage: e0.image()
            (1/(t-1))*e0 + (t/(t-1))*e1

        It is the case even if the result is integral::

            sage: x = e1.image(); x
            e1
            sage: x.parent()
            Ore module <e0, e1> over Fraction Field of Univariate Polynomial Ring in t over Rational Field twisted by d/dt

        We can force to stay in the Ore module by using the argument
        ``integral=True``::

            sage: x = e1.image(integral=True); x
            e1
            sage: x.parent()
            Ore module <e0, e1> over Univariate Polynomial Ring in t over Rational Field twisted by d/dt

        This construction however produces an error when the result is
        outside the Ore module::

            sage: e0.image(integral=True)
            Traceback (most recent call last):
            ...
            TypeError: Unable to coerce entries (=[1/(t - 1), t/(t - 1)]) to coefficients in Univariate Polynomial Ring in t over Rational Field
        """
        M = self.parent()
        y = M._pseudohom(self)
        if M._denominator is not None:
            den = M._denominator.value()
            coords = [num/den for num in y.list()]
            if not integral:
                M = M.over_fraction_field()
            y = M(coords)
        return y



# sage.doctest: needs sage.modules

from cypari2.gen cimport Gen
from cypari2.types cimport (GEN, typ, t_INT, t_FRAC, t_REAL, t_COMPLEX,
                            t_INTMOD, t_PADIC, t_INFINITY, t_VEC, t_COL,
                            t_VECSMALL, t_MAT, t_STR,
                            lg, precp)
from cypari2.paridecl cimport *

from sage.matrix.args cimport (MatrixArgs, MA_ENTRIES_SEQ_SEQ,
                               MA_ENTRIES_SEQ_FLAT, MA_ENTRIES_CALLABLE,
                               MA_ENTRIES_UNKNOWN, MA_ENTRIES_SCALAR)

from .convert_sage cimport gen_to_sage


def gen_to_sage_matrix(Gen z, locals=None):
    cdef GEN g = z.g
    nc = lg(g) - 1
    nr = 0 if nc == 0 else lg(gel(g,1)) - 1
    ma = MatrixArgs.__new__(MatrixArgs)
    ma.nrows = nr
    ma.ncols = nc
    ma.entries = [gen_to_sage(z[i,j], locals) for i in range(nr) for j in range(nc)]
    return ma.matrix()


def pari_typ_to_entries_type(MatrixArgs self):
    """
    Determine the ``entries_type`` of a :class:`sage.matrix.args.MatrixArgs`
    with PARI entries.

    This will modify the entries.

    TESTS:

    ``MA_ENTRIES_SEQ_SEQ``::

        sage: from sage.libs.pari.convert_sage import pari_typ_to_entries_type
        sage: from sage.matrix.args import MatrixArgs
        sage: ma = MatrixArgs(QQ, entries=pari("[1,2;3,4]"))
        sage: 0x10_03 == pari_typ_to_entries_type(ma)
        True

    ``MA_ENTRIES_SEQ_FLAT``::

        sage: ma = MatrixArgs(QQ, entries=pari("[1,2]"))
        sage: 0x10_04 == pari_typ_to_entries_type(ma)
        True
        sage: ma = MatrixArgs(QQ, entries=pari(vector([1,2])))
        sage: 0x10_04 == pari_typ_to_entries_type(ma)
        True
        sage: ma = MatrixArgs(QQ, entries=pari(matrix(2, range(4))[0]))
        sage: 0x10_04 == pari_typ_to_entries_type(ma)
        True

    ``MA_ENTRIES_CALLABLE``::

        sage: ma = MatrixArgs(QQ, entries=pari(lambda x: x))
        sage: 0x13_06 == pari_typ_to_entries_type(ma)
        True

    ``MA_ENTRIES_SCALAR``::

        sage: ma = MatrixArgs(QQ, entries=pari(1/2))
        sage: 0x17_02 == pari_typ_to_entries_type(ma)
        True

    ``MA_ENTRIES_UNKNOWN``::

        sage: ma = MatrixArgs(QQ, entries=pari('"2"'))
        sage: 0 == pari_typ_to_entries_type(ma)
        True

    A second call gives an error::

        sage: ma = MatrixArgs(QQ, entries=pari("[1,2]"))
        sage: 0x10_04 == pari_typ_to_entries_type(ma)
        True
        sage: 0x10_04 == pari_typ_to_entries_type(ma)
        Traceback (most recent call last):
        ...
        ValueError: entries are not a PARI generator
    """
    if not isinstance(self.entries, Gen):
        raise ValueError("entries are not a PARI generator")
    cdef long t = typ((<Gen>self.entries).g)
    if t == t_MAT:
        R = self.base
        if R is None:
            self.entries = self.entries.Col().sage()
        else:
            self.entries = [[R(x) for x in v]
                            for v in self.entries.mattranspose()]
        return MA_ENTRIES_SEQ_SEQ
    elif t in [t_VEC, t_COL, t_VECSMALL, t_LIST]:
        self.entries = self.entries.sage()
        return MA_ENTRIES_SEQ_FLAT
    elif t == t_CLOSURE:
        return MA_ENTRIES_CALLABLE
    elif t == t_STR:
        return MA_ENTRIES_UNKNOWN
    else:
        self.entries = self.entries.sage()
        return MA_ENTRIES_SCALAR



# distutils: extra_compile_args = -D_XPG6
"""
Monomial expansion of `(aX + bY)^i (cX + dY)^{j-i}`
"""

##########################################################################
#
#       Copyright (C) 2008 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
#
##########################################################################

from sage.ext.stdsage cimport PY_NEW

from sage.libs.flint.fmpz_poly cimport *
from sage.libs.flint.fmpz_poly_sage cimport *
from sage.rings.integer cimport Integer

cdef class Apply:
    def __cinit__(self):
        """
        EXAMPLES::

            sage: import sage.modular.modsym.apply
            sage: sage.modular.modsym.apply.Apply()
            <sage.modular.modsym.apply.Apply object at ...>
        """
        fmpz_poly_init(self.f)
        fmpz_poly_init(self.g)
        fmpz_poly_init(self.ff)
        fmpz_poly_init(self.gg)

    def __dealloc__(self):
        # clear flint polys
        fmpz_poly_clear(self.f)
        fmpz_poly_clear(self.g)
        fmpz_poly_clear(self.ff)
        fmpz_poly_clear(self.gg)

    cdef int apply_to_monomial_flint(self, fmpz_poly_t ans, int i, int j,
                                     int a, int b, int c, int d) except -1:
        if i < 0 or j - i < 0:
            raise ValueError("i (=%s) and j-i (=%s) must both be nonnegative." % (i, j - i))

        # f = b+a*x, g = d+c*x
        fmpz_poly_set_coeff_si(self.f, 0, b)
        fmpz_poly_set_coeff_si(self.f, 1, a)
        fmpz_poly_set_coeff_si(self.g, 0, d)
        fmpz_poly_set_coeff_si(self.g, 1, c)

        # h = (f**i)*(g**(j-i))
        fmpz_poly_pow(self.ff, self.f, i)
        fmpz_poly_pow(self.gg, self.g, j - i)
        fmpz_poly_mul(ans, self.ff, self.gg)

        return 0


cdef Apply A = Apply()


def apply_to_monomial(int i, int j, int a, int b, int c, int d):
    r"""
    Return a list of the coefficients of

    .. MATH::

        (aX + bY)^i (cX + dY)^{j-i},

    where `0 \leq i \leq j`, and `a`, `b`, `c`, `d` are integers.

    One should think of `j` as being `k-2` for the application to
    modular symbols.

    INPUT:

    - i, j, a, b, c, d -- all ints

    OUTPUT:

    list of ints, which are the coefficients
    of `Y^j, Y^{j-1}X, \ldots, X^j`, respectively.

    EXAMPLES:

    We compute that `(X+Y)^2(X-Y) = X^3 + X^2Y - XY^2 - Y^3`::

        sage: from sage.modular.modsym.apply import apply_to_monomial
        sage: apply_to_monomial(2, 3, 1,1,1,-1)
        [-1, -1, 1, 1]
        sage: apply_to_monomial(5, 8, 1,2,3,4)
        [2048, 9728, 20096, 23584, 17200, 7984, 2304, 378, 27]
        sage: apply_to_monomial(6,12, 1,1,1,-1)
        [1, 0, -6, 0, 15, 0, -20, 0, 15, 0, -6, 0, 1]
    """
    cdef fmpz_poly_t pr
    fmpz_poly_init(pr)

    A.apply_to_monomial_flint(pr, i, j, a, b, c, d)

    cdef Integer res
    v = []
    for k in range(j + 1):
        res = <Integer>PY_NEW(Integer)
        fmpz_poly_get_coeff_mpz(res.value, pr, k)
        v.append(int(res))

    fmpz_poly_clear(pr)
    return v



r"""
Hyperelliptic curves (smooth model) over the rationals

AUTHORS:

- David Kohel (2006): initial version
- Sabrina Kunzweiler, Gareth Ma, Giacomo Pope (2024): adapt to smooth model
"""

# ****************************************************************************
#       Copyright (C) 2006 David Kohel <kohel@maths.usyd.edu>
#                     2025 Sabrina Kunzweiler <sabrina.kunzweiler@math.u-bordeaux.fr>
#                     2025 Gareth Ma <grhkm21@gmail.com>
#                     2025 Giacomo Pope <giacomopope@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sage.rings.abc
from sage.rings.integer import Integer
from sage.rings.padics.factory import Qp as pAdicField
from sage.schemes.hyperelliptic_curves import hyperelliptic_generic


class HyperellipticCurve_rational_field(
    hyperelliptic_generic.HyperellipticCurve_generic
):
    def __init__(
        self, projective_model, f, h, genus: Integer, names=["x", "y"]
    ) -> None:
        r"""
        Create a hyperelliptic curve over the rationals.

        TESTS::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(-x^2, x^3 + 1)
            sage: H
            Hyperelliptic Curve over Rational Field defined by y^2 + (x^3 + 1)*y = -x^2
        """
        super().__init__(projective_model, f, h, genus, names)

    def matrix_of_frobenius(self, p, prec=20):
        r"""
        Compute the matrix of Frobenius on Monsky-Washnitzer cohomology using
        the `p`-adic field with precision ``prec``.

        This function is essentially a wrapper function of
        :meth:`sage.schemes.hyperelliptic_curves.monsky_washnitzer.matrix_of_frobenius_hyperelliptic`.

        INPUT:

        - ``p`` (prime integer or pAdic ring / field ) -- if ``p`` is an integer,
          constructs a ``pAdicField`` with ``p`` to compute the matrix of
          Frobenius, otherwise uses the supplied pAdic ring or field.

        - ``prec`` (optional) -- if ``p`` is an prime integer, the `p`-adic
          precision of the coefficient ring constructed

        EXAMPLES::

            sage: K = pAdicField(5, prec=3)
            sage: R.<x> = QQ['x']
            sage: H = HyperellipticCurve(x^5 - 2*x + 3)
            sage: H.matrix_of_frobenius(K)
            [            4*5 + O(5^3)       5 + 2*5^2 + O(5^3) 2 + 3*5 + 2*5^2 + O(5^3)     2 + 5 + 5^2 + O(5^3)]
            [      3*5 + 5^2 + O(5^3)             3*5 + O(5^3)             4*5 + O(5^3)         2 + 5^2 + O(5^3)]
            [    4*5 + 4*5^2 + O(5^3)     3*5 + 2*5^2 + O(5^3)       5 + 3*5^2 + O(5^3)     2*5 + 2*5^2 + O(5^3)]
            [            5^2 + O(5^3)       5 + 4*5^2 + O(5^3)     4*5 + 3*5^2 + O(5^3)             2*5 + O(5^3)]

        You can also pass directly a prime `p` with to construct a pAdic field with precision
        ``prec``::

            sage: H.matrix_of_frobenius(3, prec=2)
            [        O(3^2)     3 + O(3^2)         O(3^2)         O(3^2)]
            [    3 + O(3^2)         O(3^2)         O(3^2) 2 + 3 + O(3^2)]
            [  2*3 + O(3^2)         O(3^2)         O(3^2)    3^-1 + O(3)]
            [        O(3^2)         O(3^2)     3 + O(3^2)         O(3^2)]
        """
        from sage.schemes.hyperelliptic_curves import monsky_washnitzer

        if isinstance(p, (sage.rings.abc.pAdicField, sage.rings.abc.pAdicRing)):
            K = p
        else:
            K = pAdicField(p, prec)
        frob_p, _ = monsky_washnitzer.matrix_of_frobenius_hyperelliptic(
            self.change_ring(K)
        )
        return frob_p

    def lseries(self, prec=53):
        r"""
        Return the L-series of this hyperelliptic curve of genus 2.

        EXAMPLES::

            sage: x = polygen(QQ, 'x')
            sage: C = HyperellipticCurve(x^2+x, x^3+x^2+1)
            sage: C.lseries()
            PARI L-function associated to Hyperelliptic Curve
            over Rational Field defined by y^2 + (x^3 + x^2 + 1)*y = x^2 + x
        """
        from sage.lfunctions.pari import LFunction, lfun_genus2

        L = LFunction(lfun_genus2(self), prec=prec)
        L.rename("PARI L-function associated to %s" % self)
        return L



r"""
Interface with Cliquer (clique-related problems)

This module defines functions based on Cliquer, an exact
branch-and-bound algorithm developed by Patric R. J. Ostergard and
written by Sampo Niskanen.

AUTHORS:

- Nathann Cohen (2009-08-14): Initial version

- Jeroen Demeyer (2011-05-06): Make cliquer interruptible (:issue:`11252`)

- Nico Van Cleemput (2013-05-27): Handle the empty graph (:issue:`14525`)

REFERENCE:

[NO2003]_

Methods
-------
"""

# ****************************************************************************
#       Copyright (C) 2009 Nathann Cohen <nathann.cohen@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.memory cimport sig_free
from cysignals.signals cimport sig_on, sig_off


cdef extern from "cliquer/cl.c":
    cdef int sage_clique_max(graph_t *g, int ** list_of_vertices)
    cdef int sage_all_clique_max(graph_t *g, int ** list_of_vertices)
    cdef int sage_clique_number(graph_t *g)
    cdef int sage_find_all_clique(graph_t *g, int ** list_of_vertices, int min_size, int max_size)


def max_clique(graph):
    """
    Return the vertex set of a maximum complete subgraph.

    .. NOTE::

        Currently only implemented for undirected graphs. Use
        :meth:`~sage.graphs.digraph.DiGraph.to_undirected` to convert a digraph
        to an undirected graph.

    EXAMPLES::

          sage: C = graphs.PetersenGraph()
          sage: from sage.graphs.cliquer import max_clique
          sage: max_clique(C)
          [7, 9]

    TESTS::

        sage: g = Graph()
        sage: g.clique_maximum()
        []
    """
    if not graph.order():
        return []

    cdef int i
    cdef list int_to_vertex = list(graph)
    cdef dict vertex_to_int = {v: i for i, v in enumerate(int_to_vertex)}

    cdef graph_t* g = graph_new(graph.order())
    for u, v in graph.edge_iterator(labels=None):
        GRAPH_ADD_EDGE(g, vertex_to_int[u], vertex_to_int[v])

    cdef int* list_of_vertices
    cdef int size
    sig_on()
    size = sage_clique_max(g, &list_of_vertices)
    sig_off()
    cdef list b = [int_to_vertex[list_of_vertices[i]] for i in range(size)]

    sig_free(list_of_vertices)
    graph_free(g)
    return b


# computes all the maximum clique of a graph and return its list

def all_max_clique(graph):
    """
    Return the vertex sets of *ALL* the maximum complete subgraphs.

    Returns the list of all maximum cliques, with each clique represented by a
    list of vertices. A clique is an induced complete subgraph, and a maximum
    clique is one of maximal order.

    .. NOTE::

        Currently only implemented for undirected graphs. Use
        :meth:`~sage.graphs.digraph.DiGraph.to_undirected` to convert a digraph
        to an undirected graph.

    ALGORITHM:

    This function is based on Cliquer [NO2003]_.

    EXAMPLES::

        sage: graphs.ChvatalGraph().cliques_maximum() # indirect doctest
        [[0, 1], [0, 4], [0, 6], [0, 9], [1, 2], [1, 5], [1, 7], [2, 3],
         [2, 6], [2, 8], [3, 4], [3, 7], [3, 9], [4, 5], [4, 8], [5, 10],
         [5, 11], [6, 10], [6, 11], [7, 8], [7, 11], [8, 10], [9, 10], [9, 11]]
        sage: G = Graph({0:[1,2,3], 1:[2], 3:[0,1]})
        sage: G.show(figsize=[2,2])                                                     # needs sage.plot
        sage: G.cliques_maximum()
        [[0, 1, 2], [0, 1, 3]]
        sage: C = graphs.PetersenGraph()
        sage: C.cliques_maximum()
        [[0, 1], [0, 4], [0, 5], [1, 2], [1, 6], [2, 3], [2, 7], [3, 4],
         [3, 8], [4, 9], [5, 7], [5, 8], [6, 8], [6, 9], [7, 9]]
        sage: C = Graph('DJ{')
        sage: C.cliques_maximum()
        [[1, 2, 3, 4]]

    TESTS::

        sage: g = Graph()
        sage: g.cliques_maximum()
        [[]]
    """
    if not graph.order():
        return [[]]

    cdef int i
    cdef list int_to_vertex = list(graph)
    cdef dict vertex_to_int = {v: i for i, v in enumerate(int_to_vertex)}

    cdef graph_t* g = graph_new(graph.order())
    for u, v in graph.edge_iterator(labels=None):
        GRAPH_ADD_EDGE(g, vertex_to_int[u], vertex_to_int[v])

    cdef int* list_of_vertices
    cdef int size
    sig_on()
    size = sage_all_clique_max(g, &list_of_vertices)
    sig_off()
    cdef list b = []
    cdef list c = []
    for i in range(size):
        if list_of_vertices[i] != -1:
            c.append(int_to_vertex[list_of_vertices[i]])
        else:
            b.append(c)
            c = []

    sig_free(list_of_vertices)
    graph_free(g)

    return sorted(b)


def all_cliques(graph, min_size=0, max_size=0):
    r"""
    Iterator over the cliques in ``graph``.

    A clique is an induced complete subgraph. This method is an iterator over
    all the cliques with size in between ``min_size`` and ``max_size``. By
    default, this method returns only maximum cliques. Each yielded clique is
    represented by a list of vertices.

    .. NOTE::

        Currently only implemented for undirected graphs. Use
        :meth:`~sage.graphs.digraph.DiGraph.to_undirected` to convert a digraph
        to an undirected graph.

    INPUT:

    - ``min_size`` -- integer (default: 0); minimum size of reported cliques.
      When set to 0 (default), this method searches for maximum cliques. In such
      case, parameter ``max_size`` must also be set to 0.

    - ``max_size`` -- integer (default: 0); maximum size of reported cliques.
      When set to 0 (default), the maximum size of the cliques is unbounded.
      When ``min_size`` is set to 0, this parameter must be set to 0.

    ALGORITHM:

    This function is based on Cliquer [NO2003]_.

    EXAMPLES::

        sage: G = graphs.CompleteGraph(5)
        sage: list(sage.graphs.cliquer.all_cliques(G))
        [[0, 1, 2, 3, 4]]
        sage: list(sage.graphs.cliquer.all_cliques(G, 2, 3))
        [[3, 4],
         [2, 3],
         [2, 3, 4],
         [2, 4],
         [1, 2],
         [1, 2, 3],
         [1, 2, 4],
         [1, 3],
         [1, 3, 4],
         [1, 4],
         [0, 1],
         [0, 1, 2],
         [0, 1, 3],
         [0, 1, 4],
         [0, 2],
         [0, 2, 3],
         [0, 2, 4],
         [0, 3],
         [0, 3, 4],
         [0, 4]]
        sage: G.delete_edge([1,3])
        sage: list(sage.graphs.cliquer.all_cliques(G))
        [[0, 2, 3, 4], [0, 1, 2, 4]]

    TESTS::

        sage: G = graphs.CompleteGraph(3)
        sage: list(sage.graphs.cliquer.all_cliques(G, 2, 3))
        [[1, 2], [0, 1], [0, 1, 2], [0, 2]]

        sage: G = graphs.EmptyGraph()
        sage: list(sage.graphs.cliquer.all_cliques(G, 2, 3))
        []

        sage: G = Graph([(0, 1), (0, 1)], multiedges=True)
        sage: list(sage.graphs.cliquer.all_cliques(G, 2, 2))
        [[0, 1]]

        sage: list(sage.graphs.cliquer.all_cliques(G, 0, 2))
        Traceback (most recent call last):
        ...
        ValueError: max_size > 0 is incompatible with min_size == 0

    .. TODO::

        Use the re-entrant functionality of Cliquer [NO2003]_ to avoid storing
        all cliques.
    """
    if not min_size and max_size > 0:
        raise ValueError("max_size > 0 is incompatible with min_size == 0")
    if not graph:
        return

    cdef int i
    cdef list int_to_vertex = list(graph)
    cdef dict vertex_to_int = {v: i for i, v in enumerate(int_to_vertex)}

    cdef graph_t* g = graph_new(graph.order())
    for u, v in graph.edge_iterator(labels=None):
        GRAPH_ADD_EDGE(g, vertex_to_int[u], vertex_to_int[v])

    cdef int* list_of_vertices
    cdef int size = 0
    cdef list c
    try:
        try:
            sig_on()
            size = sage_find_all_clique(g, &list_of_vertices, min_size, max_size)
            sig_off()
        finally:
            graph_free(g)
        c = []
        for i in range(size):
            if list_of_vertices[i] != -1:
                c.append(int_to_vertex[list_of_vertices[i]])
            else:
                yield c
                c = []
    finally:
        if list_of_vertices:
            # We free ``list_of_vertices``,
            # but only if previous computations weren't interrupted before
            # allocating memory for ``list_of_vertices``.
            sig_free(list_of_vertices)


# computes the clique number of a graph

def clique_number(graph):
    """
    Return the size of the largest clique of the graph (clique number).

    .. NOTE::

        Currently only implemented for undirected graphs. Use
        :meth:`~sage.graphs.digraph.DiGraph.to_undirected` to convert a digraph
        to an undirected graph.

    EXAMPLES::

        sage: C = Graph('DJ{')
        sage: C.clique_number()
        4
        sage: G = Graph({0:[1,2,3], 1:[2], 3:[0,1]})
        sage: G.show(figsize=[2,2])                                                     # needs sage.plot
        sage: G.clique_number()
        3

    TESTS::

        sage: g = Graph()
        sage: g.clique_number()
        0
    """
    if not graph.order():
        return 0

    cdef int i
    cdef dict vertex_to_int = {v: i for i, v in enumerate(graph)}

    cdef graph_t* g = graph_new(graph.order())
    for u, v in graph.edge_iterator(labels=None):
        GRAPH_ADD_EDGE(g, vertex_to_int[u], vertex_to_int[v])

    cdef int c
    sig_on()
    c = sage_clique_number(g)
    graph_free(g)
    sig_off()
    return c



from sage.matroids.matroid cimport Matroid
from sage.matroids.basis_exchange_matroid cimport BasisExchangeMatroid
from sage.graphs.generic_graph_pyx cimport GenericGraph_pyx

cdef class TransversalMatroid(BasisExchangeMatroid):
    cdef dict _matching
    cdef object _sets
    cdef GenericGraph_pyx _D
    cdef list _set_labels, _sets_input, _set_labels_input

    cpdef list sets(self)
    cdef dict _translate_matching(self)
    cpdef reduce_presentation(self)
    cpdef transversal_extension(self, element=*, newset=*, sets=*)



# sage.doctest: needs sage.libs.flint
r"""
Integer factorization using FLINT

AUTHORS:

- Michael Orlitzky (2023)
"""

#*****************************************************************************
#       Copyright (C) 2023 Michael Orlitzky
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from cysignals.signals cimport sig_on, sig_off

from sage.libs.flint.fmpz cimport fmpz_t, fmpz_init, fmpz_set_mpz
from sage.libs.flint.fmpz_factor cimport *
from sage.libs.flint.fmpz_factor_sage cimport *
from sage.rings.integer cimport Integer


def factor_using_flint(Integer n, unsigned bits=0):
    r"""
    Factor the nonzero integer ``n`` using FLINT.

    This function returns a list of (factor, exponent) pairs. The
    factors will be of type ``Integer``, and the exponents will be of
    type ``int``.

    INPUT:

    - ``n`` -- a nonzero sage Integer; the number to factor
    - ``bits`` -- if nonzero, passed to ``fmpz_factor_smooth``

    OUTPUT:

    A list of ``(Integer, int)`` pairs representing the factors and
    their exponents.

    EXAMPLES::

        sage: from sage.rings.factorint_flint import factor_using_flint
        sage: n = ZZ(9962572652930382)
        sage: factors = factor_using_flint(n)
        sage: factors
        [(2, 1), (3, 1), (1660428775488397, 1)]
        sage: prod( f^e for (f,e) in factors ) == n
        True

     Negative numbers will have a leading factor of ``(-1)^1``::

        sage: n = ZZ(-1 * 2 * 3)
        sage: factor_using_flint(n)
        [(-1, 1), (2, 1), (3, 1)]

    The factorization of unity is empty::

        sage: factor_using_flint(ZZ.one())
        []

    While zero has a single factor, of... zero::

        sage: factor_using_flint(ZZ.zero())
        [(0, 1)]

    TESTS:

    Check that the integers [-10,000, 10,000] are factored correctly::

        sage: all(
        ....:   prod( f^e for (f,e) in factor_using_flint(ZZ(c*k)) ) == c*k
        ....:   for k in range(10000)
        ....:   for c in [-1, 1]
        ....: )
        True
    """
    if n.is_zero():
        return [(n, int(1))]

    cdef fmpz_t p
    fmpz_init(p)
    fmpz_set_mpz(p, (<Integer>n).value)

    cdef fmpz_factor_t factors
    fmpz_factor_init(factors)

    sig_on()
    if bits:
        fmpz_factor_smooth(factors, p, bits, 0)  # TODO make proved=* customizable
    else:
        fmpz_factor(factors, p)
    sig_off()

    pairs = fmpz_factor_to_pairlist(factors)

    fmpz_factor_clear(factors)
    return pairs
